import math
import logging
import json

from sklearn.metrics import confusion_matrix
import numpy as np

from rastervision.evaluation import ClassEvaluationItem
from rastervision.evaluation import ClassificationEvaluation

log = logging.getLogger(__name__)


def is_geojson(data):
    if isinstance(data, dict):
        return True
    else:
        try:
            json.loads(data)
            retval = True
        except ValueError:
            retval = False
        return retval


def get_class_eval_item(conf_mat, class_id, class_map):
    class_name = class_map.get_by_id(class_id).name

    if conf_mat.ravel().sum() == 0:
        return ClassEvaluationItem(None, None, None, 0, 0, class_id,
                                   class_name)

    true_pos = conf_mat[class_id, class_id]
    false_pos = conf_mat[1:, class_id].sum() - true_pos
    false_neg = conf_mat[class_id, :].sum() - true_pos
    precision = float(true_pos) / (true_pos + false_pos)
    recall = float(true_pos) / (true_pos + false_neg)
    f1 = 2 * (precision * recall) / (precision + recall)
    count_error = int(false_pos + false_neg)
    gt_count = conf_mat[class_id, :].sum()

    if math.isnan(precision):
        precision = None
    else:
        precision = float(precision)
    if math.isnan(recall):
        recall = None
    else:
        recall = float(recall)
    if math.isnan(f1):
        f1 = None
    else:
        f1 = float(f1)

    return ClassEvaluationItem(precision, recall, f1, count_error, gt_count,
                               class_id, class_name, conf_mat[class_id, :])


class SemanticSegmentationEvaluation(ClassificationEvaluation):
    """Evaluation for semantic segmentation."""

    def __init__(self, class_map):
        super().__init__()
        self.class_map = class_map

    def compute(self, gt_labels, pred_labels):
        self.clear()

        # Ensure there is a row and column for each class_id even when
        # class_ids are non-consecutive.
        labels = np.arange(max(self.class_map.get_keys()) + 1)
        conf_mat = np.zeros((len(labels), len(labels)))
        for window in pred_labels.get_windows():
            log.debug('Evaluating window: {}'.format(window))
            gt_arr = gt_labels.get_label_arr(window).ravel()
            pred_arr = pred_labels.get_label_arr(window).ravel()
            conf_mat += confusion_matrix(gt_arr, pred_arr, labels=labels)

        for class_id in self.class_map.get_keys():
            if class_id > 0:
                self.class_to_eval_item[class_id] = get_class_eval_item(
                    conf_mat, class_id, self.class_map)
        self.compute_avg()

    def compute_vector(self, gt, pred, mode, class_id):
        """Compute evaluation over vector predictions.
            Args:
                gt: Ground-truth GeoJSON.  Either a string (containing
                    unparsed GeoJSON or a file name), or a dictionary
                    containing parsed GeoJSON.
                pred: GeoJSON for predictions.  Either a string
                    (containing unparsed GeoJSON or a file name), or a
                    dictionary containing parsed GeoJSON.
                mode: A string containing either 'buildings' or
                    'polygons'.
                class_id: An integer containing the class id of
                    interest.
        """
        import mask_to_polygons.vectorification as vectorification
        import mask_to_polygons.processing.score as score

        # Ground truth as list of geometries
        def get_geoms(x):
            if is_geojson(x):
                _x = x
                if 'features' in _x.keys():
                    _x = _x['features']
                geoms = []
                for feature in _x:
                    if 'geometry' in feature.keys():
                        geoms.append(feature['geometry'])
                    else:
                        geoms.append(feature)
            else:
                geoms = vectorification.geometries_from_geojson(x)

            return geoms

        gt = get_geoms(gt)
        pred = get_geoms(pred)

        if len(gt) > 0 and len(pred) > 0:
            results = score.spacenet(pred, gt)

            true_positives = results['tp']
            false_positives = results['fp']
            false_negatives = results['fn']
            precision = float(true_positives) / (
                true_positives + false_positives)
            recall = float(true_positives) / (true_positives + false_negatives)
            if precision + recall != 0:
                f1 = 2 * (precision * recall) / (precision + recall)
            else:
                f1 = 0.0
            count_error = int(false_positives + false_negatives)
            gt_count = len(gt)
            class_name = 'vector-{}-{}'.format(
                mode,
                self.class_map.get_by_id(class_id).name)

            evaluation_item = ClassEvaluationItem(precision, recall, f1,
                                                  count_error, gt_count,
                                                  class_id, class_name)

            if hasattr(self, 'class_to_eval_item') and isinstance(
                    self.class_to_eval_item, dict):
                self.class_to_eval_item[class_id] = evaluation_item
            else:
                self.class_to_eval_item = {class_id: evaluation_item}
            self.compute_avg()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/class_item.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/class_item.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n$rastervision/protos/class_item.proto\x12\trv.protos\"4\n\tClassItem\x12\n\n\x02id\x18\x01 \x02(\x05\x12\x0c\n\x04name\x18\x02 \x02(\t\x12\r\n\x05\x63olor\x18\x03 \x01(\t')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_CLASSITEM = _descriptor.Descriptor(
  name='ClassItem',
  full_name='rv.protos.ClassItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='rv.protos.ClassItem.id', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='rv.protos.ClassItem.name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='color', full_name='rv.protos.ClassItem.color', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=51,
  serialized_end=103,
)

DESCRIPTOR.message_types_by_name['ClassItem'] = _CLASSITEM

ClassItem = _reflection.GeneratedProtocolMessageType('ClassItem', (_message.Message,), dict(
  DESCRIPTOR = _CLASSITEM,
  __module__ = 'rastervision.protos.class_item_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.ClassItem)
  ))
_sym_db.RegisterMessage(ClassItem)


# @@protoc_insertion_point(module_scope)



# flake8: noqa

# Registry Keys
LABEL_STORE = 'LABEL_STORE'

# Use the same labels as for the source for these label stores.
from rastervision.data.label_source.api import (OBJECT_DETECTION_GEOJSON,
                                                CHIP_CLASSIFICATION_GEOJSON,
                                                SEMANTIC_SEGMENTATION_RASTER)

from rastervision.data.label_store.label_store_config import LabelStoreConfig



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/argmax_matcher.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/argmax_matcher.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\n<rastervision/protos/tf_object_detection/argmax_matcher.proto\x12\'rastervision.protos.tf_object_detection\"\xec\x01\n\rArgMaxMatcher\x12\x1e\n\x11matched_threshold\x18\x01 \x01(\x02:\x03\x30.5\x12 \n\x13unmatched_threshold\x18\x02 \x01(\x02:\x03\x30.5\x12 \n\x11ignore_thresholds\x18\x03 \x01(\x08:\x05\x66\x61lse\x12,\n\x1enegatives_lower_than_unmatched\x18\x04 \x01(\x08:\x04true\x12\'\n\x18\x66orce_match_for_each_row\x18\x05 \x01(\x08:\x05\x66\x61lse\x12 \n\x11use_matmul_gather\x18\x06 \x01(\x08:\x05\x66\x61lse')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_ARGMAXMATCHER = _descriptor.Descriptor(
  name='ArgMaxMatcher',
  full_name='rastervision.protos.tf_object_detection.ArgMaxMatcher',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='matched_threshold', full_name='rastervision.protos.tf_object_detection.ArgMaxMatcher.matched_threshold', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='unmatched_threshold', full_name='rastervision.protos.tf_object_detection.ArgMaxMatcher.unmatched_threshold', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0.5),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ignore_thresholds', full_name='rastervision.protos.tf_object_detection.ArgMaxMatcher.ignore_thresholds', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='negatives_lower_than_unmatched', full_name='rastervision.protos.tf_object_detection.ArgMaxMatcher.negatives_lower_than_unmatched', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='force_match_for_each_row', full_name='rastervision.protos.tf_object_detection.ArgMaxMatcher.force_match_for_each_row', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='use_matmul_gather', full_name='rastervision.protos.tf_object_detection.ArgMaxMatcher.use_matmul_gather', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=106,
  serialized_end=342,
)

DESCRIPTOR.message_types_by_name['ArgMaxMatcher'] = _ARGMAXMATCHER

ArgMaxMatcher = _reflection.GeneratedProtocolMessageType('ArgMaxMatcher', (_message.Message,), dict(
  DESCRIPTOR = _ARGMAXMATCHER,
  __module__ = 'rastervision.protos.tf_object_detection.argmax_matcher_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.ArgMaxMatcher)
  ))
_sym_db.RegisterMessage(ArgMaxMatcher)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/keras_classification/trainer.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rastervision.protos.keras_classification import optimizer_pb2 as rastervision_dot_protos_dot_keras__classification_dot_optimizer__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/keras_classification/trainer.proto',
  package='keras_classification.protos',
  syntax='proto2',
  serialized_pb=_b('\n6rastervision/protos/keras_classification/trainer.proto\x12\x1bkeras_classification.protos\x1a\x38rastervision/protos/keras_classification/optimizer.proto\"\xe6\x03\n\x07Trainer\x12\x39\n\toptimizer\x18\x01 \x02(\x0b\x32&.keras_classification.protos.Optimizer\x12=\n\x07options\x18\x02 \x02(\x0b\x32,.keras_classification.protos.Trainer.Options\x1a+\n\x0eLRScheduleItem\x12\r\n\x05\x65poch\x18\x01 \x02(\x05\x12\n\n\x02lr\x18\x02 \x02(\x02\x1a\xb3\x02\n\x07Options\x12\x19\n\x11training_data_dir\x18\x01 \x02(\t\x12\x1b\n\x13validation_data_dir\x18\x02 \x02(\t\x12\x11\n\tnb_epochs\x18\x04 \x02(\x05\x12H\n\x0blr_schedule\x18\n \x03(\x0b\x32\x33.keras_classification.protos.Trainer.LRScheduleItem\x12\x12\n\nbatch_size\x18\x05 \x02(\x05\x12\x12\n\ninput_size\x18\x06 \x02(\x05\x12\x12\n\noutput_dir\x18\x07 \x02(\t\x12\x13\n\x0b\x63lass_names\x18\x08 \x03(\t\x12\x1a\n\x0bshort_epoch\x18\t \x02(\x08:\x05\x66\x61lse\x12\r\n\x05\x64\x65\x62ug\x18\x0b \x02(\x08\x12\x17\n\tsave_best\x18\x0c \x02(\x08:\x04true')
  ,
  dependencies=[rastervision_dot_protos_dot_keras__classification_dot_optimizer__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_TRAINER_LRSCHEDULEITEM = _descriptor.Descriptor(
  name='LRScheduleItem',
  full_name='keras_classification.protos.Trainer.LRScheduleItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='epoch', full_name='keras_classification.protos.Trainer.LRScheduleItem.epoch', index=0,
      number=1, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='lr', full_name='keras_classification.protos.Trainer.LRScheduleItem.lr', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=279,
  serialized_end=322,
)

_TRAINER_OPTIONS = _descriptor.Descriptor(
  name='Options',
  full_name='keras_classification.protos.Trainer.Options',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='training_data_dir', full_name='keras_classification.protos.Trainer.Options.training_data_dir', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='validation_data_dir', full_name='keras_classification.protos.Trainer.Options.validation_data_dir', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='nb_epochs', full_name='keras_classification.protos.Trainer.Options.nb_epochs', index=2,
      number=4, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='lr_schedule', full_name='keras_classification.protos.Trainer.Options.lr_schedule', index=3,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='batch_size', full_name='keras_classification.protos.Trainer.Options.batch_size', index=4,
      number=5, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='input_size', full_name='keras_classification.protos.Trainer.Options.input_size', index=5,
      number=6, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='output_dir', full_name='keras_classification.protos.Trainer.Options.output_dir', index=6,
      number=7, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='class_names', full_name='keras_classification.protos.Trainer.Options.class_names', index=7,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='short_epoch', full_name='keras_classification.protos.Trainer.Options.short_epoch', index=8,
      number=9, type=8, cpp_type=7, label=2,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='debug', full_name='keras_classification.protos.Trainer.Options.debug', index=9,
      number=11, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='save_best', full_name='keras_classification.protos.Trainer.Options.save_best', index=10,
      number=12, type=8, cpp_type=7, label=2,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=325,
  serialized_end=632,
)

_TRAINER = _descriptor.Descriptor(
  name='Trainer',
  full_name='keras_classification.protos.Trainer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='optimizer', full_name='keras_classification.protos.Trainer.optimizer', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='options', full_name='keras_classification.protos.Trainer.options', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_TRAINER_LRSCHEDULEITEM, _TRAINER_OPTIONS, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=146,
  serialized_end=632,
)

_TRAINER_LRSCHEDULEITEM.containing_type = _TRAINER
_TRAINER_OPTIONS.fields_by_name['lr_schedule'].message_type = _TRAINER_LRSCHEDULEITEM
_TRAINER_OPTIONS.containing_type = _TRAINER
_TRAINER.fields_by_name['optimizer'].message_type = rastervision_dot_protos_dot_keras__classification_dot_optimizer__pb2._OPTIMIZER
_TRAINER.fields_by_name['options'].message_type = _TRAINER_OPTIONS
DESCRIPTOR.message_types_by_name['Trainer'] = _TRAINER

Trainer = _reflection.GeneratedProtocolMessageType('Trainer', (_message.Message,), dict(

  LRScheduleItem = _reflection.GeneratedProtocolMessageType('LRScheduleItem', (_message.Message,), dict(
    DESCRIPTOR = _TRAINER_LRSCHEDULEITEM,
    __module__ = 'rastervision.protos.keras_classification.trainer_pb2'
    # @@protoc_insertion_point(class_scope:keras_classification.protos.Trainer.LRScheduleItem)
    ))
  ,

  Options = _reflection.GeneratedProtocolMessageType('Options', (_message.Message,), dict(
    DESCRIPTOR = _TRAINER_OPTIONS,
    __module__ = 'rastervision.protos.keras_classification.trainer_pb2'
    # @@protoc_insertion_point(class_scope:keras_classification.protos.Trainer.Options)
    ))
  ,
  DESCRIPTOR = _TRAINER,
  __module__ = 'rastervision.protos.keras_classification.trainer_pb2'
  # @@protoc_insertion_point(class_scope:keras_classification.protos.Trainer)
  ))
_sym_db.RegisterMessage(Trainer)
_sym_db.RegisterMessage(Trainer.LRScheduleItem)
_sym_db.RegisterMessage(Trainer.Options)


# @@protoc_insertion_point(module_scope)



import os
import json
import importlib

from pluginbase import PluginBase

import rastervision as rv
from rastervision.protos.plugin_pb2 import PluginConfig as PluginConfigMsg
from rastervision.utils.files import download_if_needed


class PluginError(Exception):
    pass


def load_conf_list(s):
    """Loads a list of items from the config.

    Lists should be comma separated.

    This takes into account that previous versions of Raster Vision
    allowed for a `[ "module" ]` like syntax, even though that didn't
    work for multi-value lists.
    """
    try:
        # A comma separated list of values will be transformed to
        # having a list-like string, with ' instead of ". Replacing
        # single quotes with double quotes lets us parse it as a JSON list.
        return json.loads(s.replace("'", '"'))
    except json.JSONDecodeError:
        return list(map(lambda x: x.strip(), s.split(',')))


class PluginRegistry:
    @staticmethod
    def get_instance():
        return rv._registry._get_plugin_registry()

    def __init__(self, plugin_config, rv_home):
        """Initializes this plugin registry.

        A plugin registry is passed to plugins in a call
        to their "register_plugin" method.

        Args:
           plugin_config - the everett ConfigManager for the plugin
                           section of the application configuration.
        """
        self.plugin_root_dir = os.path.join(rv_home, 'plugins')
        self.config_builders = {}
        self.command_config_builders = {}
        self.commands = []
        self.aux_command_classes = {}
        self.default_raster_sources = []
        self.default_vector_sources = []
        self.default_label_sources = []
        self.default_label_stores = []
        self.default_evaluators = []
        self.experiment_runners = {}
        self.filesystems = []

        plugin_files = load_conf_list(plugin_config('files', default='[]'))
        self._load_from_files(plugin_files)
        self.plugin_files = plugin_files

        plugin_modules = load_conf_list(plugin_config('modules', default='[]'))
        self._load_from_modules(plugin_modules)
        self.plugin_modules = plugin_modules

    def _load_plugin(self, plugin, identifier):
        # Check the plugin is valid
        if not hasattr(plugin, 'register_plugin'):
            raise PluginError('Plugin at {} does not have '
                              '"register_plugin" method.'.format(identifier))

        register_method = getattr(plugin, 'register_plugin')
        if not callable(register_method):
            raise PluginError('Plugin at {} has a '
                              '"register_plugin" attribute, '
                              'but it is not callable'.format(identifier))

        # TODO: Log loading plugin.
        register_method(self)

    def _load_from_files(self, plugin_paths):
        if not plugin_paths:
            return

        self.plugin_sources = []

        plugin_base = PluginBase(package='rastervision.plugins')
        for uri in plugin_paths:
            plugin_name = os.path.splitext(os.path.basename(uri))[0]
            plugin_path = os.path.join(self.plugin_root_dir, plugin_name)
            fs = rv._registry.get_file_system(uri, search_plugins=False)
            local_path = download_if_needed(uri, plugin_path, fs=fs)
            local_dir = os.path.dirname(local_path)

            plugin_source = plugin_base.make_plugin_source(
                searchpath=[local_dir])

            # We're required to hang onto the source
            # to keep it from getting GC'd.
            self.plugin_sources.append(plugin_source)

            self._load_plugin(plugin_source.load_plugin(plugin_name), uri)

    def _load_from_modules(self, plugin_modules):
        if not plugin_modules:
            return

        for module in plugin_modules:
            plugin = importlib.import_module(module)
            self._load_plugin(plugin, module)

    def add_plugins_from_proto(self, plugin_msg):
        new_plugin_files = list(
            set(plugin_msg.plugin_uris) - set(self.plugin_files))
        self._load_from_files(new_plugin_files)
        self.plugin_files.extend(new_plugin_files)

        new_plugin_modules = list(
            set(plugin_msg.plugin_modules) - set(self.plugin_modules))
        self._load_from_modules(new_plugin_modules)
        self.plugin_modules.extend(new_plugin_modules)

    def to_proto(self):
        """Returns a protobuf message that records the
        plugin sources for plugins that are currently loaded
        in the registry.
        """
        return PluginConfigMsg(
            plugin_uris=self.plugin_files, plugin_modules=self.plugin_modules)

    def register_config_builder(self, group, key, builder_class):
        """Registers a ConfigBuilder as a plugin.

        Args:
           group - The Config group, e.g. rv.BACKEND, rv.TASK.
           key - The key used for this plugin. This will be used to
                 construct the builder in a ".builder(key)" call.
           builder_class - The subclass of ConfigBuilder that builds
                           the Config for this plugin.
        """
        if (group, key) in self.config_builders:
            raise PluginError('ConfigBuilder already registered for group '
                              '{} and key {}'.format(group, key))
        self.config_builders[(group, key)] = builder_class

    def register_command_config_builder(self, command_type, builder_class):
        """Registers a ConfigBuilder as a plugin.

        Args:
           command_type - The key used for this plugin. This will be used to
                          construct the builder in a ".builder(key)" call.
           builder_class - The subclass of CommandConfigBuilder that builds
                           the CommandConfig for this plugin.
        """
        if command_type in self.command_config_builders:
            raise PluginError(
                'CommandConfigBuilder already registered for command'
                'with type {}'.format(command_type))
        self.command_config_builders[command_type] = builder_class
        self.commands.append(command_type)

    def register_aux_command(self, command_type, command_class):
        """Registers a custom AuxCommand as a plugin.

        Args:
           command_type - The key used for this plugin. This will be used to
                          construct the builder in a ".builder(key)" call.
           command_class - The subclass of AuxCommand subclass to register.
        """
        if command_type in self.command_config_builders:
            raise PluginError(
                'CommandConfigBuilder is already registered for command'
                'with type {}'.format(command_type))
        if command_type in self.aux_command_classes:
            raise PluginError('AuxCommand is already registered for command'
                              'with type {}'.format(command_type))
        self.aux_command_classes[command_type] = command_class
        if command_class.options.include_by_default:
            self.commands.append(command_type)

    def register_default_raster_source(self, provider_class):
        """Registers a RasterSourceDefaultProvider for use as a plugin."""

        self.default_raster_sources.append(provider_class)

    def register_default_vector_source(self, provider_class):
        """Registers a VectorSourceDefaultProvider for use as a plugin."""
        self.default_vector_sources.append(provider_class)

    def register_default_label_source(self, provider_class):
        """Registers a LabelSourceDefaultProvider for use as a plugin."""
        self.default_label_sources.append(provider_class)

    def register_default_label_store(self, provider_class):
        """Registers a LabelStoreDefaultProvider for use as a plugin."""
        self.default_label_stores.append(provider_class)

    def register_default_evaluator(self, provider_class):
        """Registers an EvaluatorDefaultProvider for use as a plugin."""
        self.default_evaluators.append(provider_class)

    def register_experiment_runner(self, runner_key, runner_class):
        """Registers an ExperimentRunner as a plugin.

        Args:
           runner_key - The key used to reference this plugin runner.
                        This is a string that will match the command line
                        argument used to reference this runner; e.g. if the
                        key is "FOO_RUNNER", then users can use the runner
                        by issuing a "rastervision run  foo_runner ..." command.
           runner_class - The class of the ExperimentRunner plugin.
        """
        if runner_key in self.experiment_runners:
            raise PluginError('ExperimentRunner already registered for '
                              'key {}'.format(runner_key))
        self.experiment_runners[runner_key] = runner_class

    def register_filesystem(self, filesystem_class):
        """Registers a FileSystem as a plugin."""
        self.filesystems.append(filesystem_class)



from copy import deepcopy

import rastervision as rv
from rastervision.data.label_source import (
    LabelSourceConfig, LabelSourceConfigBuilder, ChipClassificationLabelSource)
from rastervision.data.vector_source import VectorSourceConfig
from rastervision.protos.label_source_pb2 import (LabelSourceConfig as
                                                  LabelSourceConfigMsg)


class ChipClassificationLabelSourceConfig(LabelSourceConfig):
    def __init__(self,
                 vector_source,
                 ioa_thresh=None,
                 use_intersection_over_cell=False,
                 pick_min_class_id=False,
                 background_class_id=None,
                 cell_size=None,
                 infer_cells=False):
        super().__init__(source_type=rv.CHIP_CLASSIFICATION)
        self.vector_source = vector_source
        self.ioa_thresh = ioa_thresh
        self.use_intersection_over_cell = use_intersection_over_cell
        self.pick_min_class_id = pick_min_class_id
        self.background_class_id = background_class_id
        self.cell_size = cell_size
        self.infer_cells = infer_cells

    def to_proto(self):
        msg = super().to_proto()
        options = LabelSourceConfigMsg.ChipClassificationLabelSource(
            vector_source=self.vector_source.to_proto(),
            ioa_thresh=self.ioa_thresh,
            use_intersection_over_cell=self.use_intersection_over_cell,
            pick_min_class_id=self.pick_min_class_id,
            background_class_id=self.background_class_id,
            cell_size=self.cell_size,
            infer_cells=self.infer_cells)
        msg.chip_classification_label_source.CopyFrom(options)
        return msg

    def create_source(self, task_config, extent, crs_transformer, tmp_dir):
        vector_source = self.vector_source.create_source(
            crs_transformer=crs_transformer,
            extent=extent,
            class_map=task_config.class_map)
        return ChipClassificationLabelSource(
            vector_source, crs_transformer, task_config.class_map, extent,
            self.ioa_thresh, self.use_intersection_over_cell,
            self.pick_min_class_id, self.background_class_id, self.cell_size,
            self.infer_cells)

    def update_for_command(self, command_type, experiment_config,
                           context=None):
        self.vector_source.update_for_command(command_type, experiment_config,
                                              context)

        if not self.cell_size:
            self.cell_size = experiment_config.task.chip_size

    def report_io(self, command_type, io_def):
        self.vector_source.report_io(command_type, io_def)


class ChipClassificationLabelSourceConfigBuilder(LabelSourceConfigBuilder):
    def __init__(self, prev=None):
        config = {}
        if prev:
            config = {
                'vector_source': prev.vector_source,
                'ioa_thresh': prev.ioa_thresh,
                'use_intersection_over_cell': prev.use_intersection_over_cell,
                'pick_min_class_id': prev.pick_min_class_id,
                'background_class_id': prev.background_class_id,
                'cell_size': prev.cell_size,
                'infer_cells': prev.infer_cells
            }

        super().__init__(ChipClassificationLabelSourceConfig, config)

    def validate(self):
        super().validate()

        vector_source = self.config.get('vector_source')
        if vector_source is None:
            raise rv.ConfigError(
                'You must set the vector_source for ChipClassificationLabelSourceConfig'
                ' Use "with_vector_source".')
        if not isinstance(vector_source, VectorSourceConfig):
            raise rv.ConfigError(
                'vector source must be a child of class VectorSourceConfig, got {}'.
                format(type(vector_source)))
        if vector_source.has_null_class_bufs():
            raise rv.ConfigError(
                'Setting buffer to None for a class in the vector_source is not allowed '
                'for ChipClassificationLabelSourceConfig.')

    def from_proto(self, msg):
        # Added for backwards compatibility.
        if msg.HasField('chip_classification_geojson_source'):
            conf = msg.chip_classification_geojson_source
            vector_source = conf.uri
        else:
            conf = msg.chip_classification_label_source
            vector_source = rv.VectorSourceConfig.from_proto(
                conf.vector_source)

        return self \
            .with_vector_source(vector_source) \
            .with_ioa_thresh(conf.ioa_thresh) \
            .with_use_intersection_over_cell(conf.use_intersection_over_cell) \
            .with_pick_min_class_id(conf.pick_min_class_id) \
            .with_background_class_id(conf.background_class_id) \
            .with_cell_size(conf.cell_size) \
            .with_infer_cells(conf.infer_cells)

    def with_vector_source(self, vector_source):
        """Set the vector_source.

        Args:
            vector_source (str or VectorSource) if a string, assume it is
                a URI and use the default provider to construct a VectorSource.
        """
        if isinstance(vector_source, str):
            return self.with_uri(vector_source)

        b = deepcopy(self)
        if isinstance(vector_source, VectorSourceConfig):
            b.config['vector_source'] = vector_source
        else:
            raise rv.ConfigError(
                'vector_source must be of type str or VectorSource')

        return b

    def with_uri(self, uri):
        b = deepcopy(self)
        provider = rv._registry.get_vector_source_default_provider(uri)
        b.config['vector_source'] = provider.construct(uri)
        return b

    def with_ioa_thresh(self, ioa_thresh):
        """The minimum IOA of a polygon and cell."""
        b = deepcopy(self)
        b.config['ioa_thresh'] = ioa_thresh
        return b

    def with_use_intersection_over_cell(self, use_intersection_over_cell):
        """ Set this label source to use intersection over cell or not.

        If use_intersection_over_cell is true, then use the area of the
        cell as the denominator in the IOA. Otherwise, use the area of the
        polygon.
        """
        b = deepcopy(self)
        b.config['use_intersection_over_cell'] = use_intersection_over_cell
        return b

    def with_pick_min_class_id(self, pick_min_class_id):
        """Set this label source to pick min class ID

        If true, the class_id for a cell is the minimum class_id of the
        boxes in that cell. Otherwise, pick the class_id of the box
        covering the greatest area.
        """
        b = deepcopy(self)
        b.config['pick_min_class_id'] = pick_min_class_id
        return b

    def with_background_class_id(self, background_class_id):
        """Sets the background class ID.

        Optional class_id to use as the background class; ie. the one that
        is used when a window contains no boxes. If not set, empty windows
        have None set as their class_id.
        """
        b = deepcopy(self)
        b.config['background_class_id'] = background_class_id
        return b

    def with_infer_cells(self, infer_cells):
        """Set if this label source should infer cells.

        If true, the label source will infer the cell polygon and label
        from the polygons in the vector_source. If the labels are already
        cells and properly labeled, this can be False.
        """
        b = deepcopy(self)
        b.config['infer_cells'] = infer_cells
        return b

    def with_cell_size(self, cell_size):
        """Sets the cell size of the chips.

        If not explicitly set, the chip size will be used if this object is
        created as part of an experiment.

        Args:
            cell_size: (int) the size of the cells in units of pixels
        """
        b = deepcopy(self)
        b.config['cell_size'] = cell_size
        return b



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/raster_transformer.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/raster_transformer.proto',
  package='rv.protos',
  syntax='proto2',
  serialized_pb=_b('\n,rastervision/protos/raster_transformer.proto\x12\trv.protos\x1a\x1cgoogle/protobuf/struct.proto\"\x97\x01\n\x17RasterTransformerConfig\x12\x18\n\x10transformer_type\x18\x01 \x02(\t\x12\x13\n\tstats_uri\x18\x04 \x01(\tH\x00\x12\x30\n\rcustom_config\x18\x05 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x42\x1b\n\x19raster_transformer_config')
  ,
  dependencies=[google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_RASTERTRANSFORMERCONFIG = _descriptor.Descriptor(
  name='RasterTransformerConfig',
  full_name='rv.protos.RasterTransformerConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transformer_type', full_name='rv.protos.RasterTransformerConfig.transformer_type', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stats_uri', full_name='rv.protos.RasterTransformerConfig.stats_uri', index=1,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='custom_config', full_name='rv.protos.RasterTransformerConfig.custom_config', index=2,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='raster_transformer_config', full_name='rv.protos.RasterTransformerConfig.raster_transformer_config',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=90,
  serialized_end=241,
)

_RASTERTRANSFORMERCONFIG.fields_by_name['custom_config'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_RASTERTRANSFORMERCONFIG.oneofs_by_name['raster_transformer_config'].fields.append(
  _RASTERTRANSFORMERCONFIG.fields_by_name['stats_uri'])
_RASTERTRANSFORMERCONFIG.fields_by_name['stats_uri'].containing_oneof = _RASTERTRANSFORMERCONFIG.oneofs_by_name['raster_transformer_config']
_RASTERTRANSFORMERCONFIG.oneofs_by_name['raster_transformer_config'].fields.append(
  _RASTERTRANSFORMERCONFIG.fields_by_name['custom_config'])
_RASTERTRANSFORMERCONFIG.fields_by_name['custom_config'].containing_oneof = _RASTERTRANSFORMERCONFIG.oneofs_by_name['raster_transformer_config']
DESCRIPTOR.message_types_by_name['RasterTransformerConfig'] = _RASTERTRANSFORMERCONFIG

RasterTransformerConfig = _reflection.GeneratedProtocolMessageType('RasterTransformerConfig', (_message.Message,), dict(
  DESCRIPTOR = _RASTERTRANSFORMERCONFIG,
  __module__ = 'rastervision.protos.raster_transformer_pb2'
  # @@protoc_insertion_point(class_scope:rv.protos.RasterTransformerConfig)
  ))
_sym_db.RegisterMessage(RasterTransformerConfig)


# @@protoc_insertion_point(module_scope)



# flake8: noqa

# Registry keys

BACKEND = 'BACKEND'

## Backend Keys

TF_OBJECT_DETECTION = 'TF_OBJECT_DETECTION'
KERAS_CLASSIFICATION = 'KERAS_CLASSIFICATION'
TF_DEEPLAB = 'TF_DEEPLAB'

PYTORCH_SEMANTIC_SEGMENTATION = 'PYTORCH_SEMANTIC_SEGMENTATION'
PYTORCH_CHIP_CLASSIFICATION = 'PYTORCH_CHIP_CLASSIFICATION'
PYTORCH_OBJECT_DETECTION = 'PYTORCH_OBJECT_DETECTION'

## Model keys

### TF Object Detection
SSD_MOBILENET_V1_COCO = 'SSD_MOBILENET_V1_COCO'
SSD_MOBILENET_V2_COCO = 'SSD_MOBILENET_V2_COCO'
SSDLITE_MOBILENET_V2_COCO = 'SSDLITE_MOBILENET_V2_COCO'
SSD_INCEPTION_V2_COCO = 'SSD_INCEPTION_V2_COCO'
FASTER_RCNN_INCEPTION_V2_COCO = 'FASTER_RCNN_INCEPTION_V2_COCO'
FASTER_RCNN_RESNET50_COCO = 'FASTER_RCNN_RESNET50_COCO'
RFCN_RESNET101_COCO = 'RFCN_RESNET101_COCO'
FASTER_RCNN_RESNET101_COCO = 'FASTER_RCNN_RESNET101_COCO'
FASTER_RCNN_INCEPTION_RESNET_V2_ATROUS_COCO = \
'FASTER_RCNN_INCEPTION_RESNET_V2_ATROUS_COCO'
MASK_RCNN_INCEPTION_RESNET_V2_ATROUS_COCO = \
'MASK_RCNN_INCEPTION_RESNET_V2_ATROUS_COCO'
MASK_RCNN_INCEPTION_V2_COCO = 'MASK_RCNN_INCEPTION_V2_COCO'
MASK_RCNN_RESNET101_ATROUS_COCO = 'MASK_RCNN_RESNET101_ATROUS_COCO'
MASK_RCNN_RESNET50_ATROUS_COCO = 'MASK_RCNN_RESNET50_ATROUS_COCO'

## Keras Classificaiton
RESNET50_IMAGENET = 'RESNET50_IMAGENET'

# TF Deeplab
XCEPTION_65 = 'XCEPTION_65'
MOBILENET_V2 = 'MOBILENET_V2'

from .backend_config import BackendConfig



import os
import zipfile
import logging

import click
from google.protobuf import json_format

from rastervision.command import Command
from rastervision.utils.files import (upload_or_copy, make_dir)

log = logging.getLogger(__name__)


class BundleCommand(Command):
    """Bundles all the necessary files together into a prediction package."""

    def __init__(self, command_config):
        self.command_config = command_config

    def run(self, tmp_dir=None):
        if not tmp_dir:
            tmp_dir = self.get_tmp_dir()

        cc = self.command_config

        if not cc.task.predict_package_uri:
            msg = 'Skipping bundling of prediction package, no URI is set...'.format(
                cc.task.predict_package_uri)
            click.echo(click.style(msg, fg='yellow'))
            return

        msg = 'Bundling prediction package to {}...'.format(
            cc.task.predict_package_uri)
        log.info(msg)

        bundle_dir = os.path.join(tmp_dir, 'bundle')
        make_dir(bundle_dir)
        package_path = os.path.join(tmp_dir, 'predict_package.zip')
        bundle_files = []
        new_task, task_files = cc.task.save_bundle_files(bundle_dir)
        bundle_files.extend(task_files)
        new_backend, backend_files = cc.backend.save_bundle_files(bundle_dir)
        bundle_files.extend(backend_files)
        new_scene, scene_files = cc.scene.save_bundle_files(bundle_dir)
        bundle_files.extend(scene_files)
        new_analyzers = []
        for analyzer in cc.analyzers:
            new_analyzer, analyzer_files = analyzer.save_bundle_files(
                bundle_dir)
            new_analyzers.append(new_analyzer)
            bundle_files.extend(analyzer_files)

        new_bundle_config = cc.to_builder() \
                              .with_task(new_task) \
                              .with_backend(new_backend) \
                              .with_scene(new_scene) \
                              .with_analyzers(new_analyzers) \
                              .build()

        # Save bundle command config
        bundle_config_path = os.path.join(tmp_dir, 'bundle_config.json')
        bundle_json = json_format.MessageToJson(new_bundle_config.to_proto())
        with open(bundle_config_path, 'w') as f:
            f.write(bundle_json)

        with zipfile.ZipFile(package_path, 'w') as package_zip:
            for path in bundle_files:
                package_zip.write(path, arcname=os.path.basename(path))
            package_zip.write(
                bundle_config_path,
                arcname=os.path.basename(bundle_config_path))

        upload_or_copy(package_path, cc.task.predict_package_uri)



import click
import torch

from rastervision.backend.torch_utils.metrics import (compute_conf_mat_metrics)


def train_epoch(model, device, data_loader, opt, loss_fn, step_scheduler=None):
    model.train()
    total_loss = 0.0
    num_samples = 0

    with click.progressbar(data_loader, label='Training') as bar:
        for batch_ind, (x, y) in enumerate(bar):
            x = x.to(device)
            y = y.to(device)

            opt.zero_grad()
            out = model(x)
            loss = loss_fn(out, y)
            loss.backward()
            total_loss += loss.item()
            opt.step()
            if step_scheduler:
                step_scheduler.step()
            num_samples += x.shape[0]

    return total_loss / num_samples


def validate_epoch(model, device, data_loader, num_labels):
    model.eval()

    conf_mat = torch.zeros((num_labels, num_labels))
    with torch.no_grad():
        with click.progressbar(data_loader, label='Validating') as bar:
            for batch_ind, (x, y) in enumerate(bar):
                x = x.to(device)
                out = model(x)

                out = out.argmax(-1).view(-1).cpu()
                y = y.cpu()
                if batch_ind == 0:
                    labels = torch.arange(0, num_labels)

                conf_mat += ((out == labels[:, None]) &
                             (y == labels[:, None, None])).sum(
                                 dim=2, dtype=torch.float32)

    return compute_conf_mat_metrics(conf_mat)



from abc import abstractmethod


class ActivationError(Exception):
    pass


class ActivateMixin:
    """Defines a mixin for data that can activate and deactivate.
    These methods can open and close files, download files, and do
    whatever has to be done to make the entity usable, and cleanup
    after the entity is not needed anymore.
    """

    class ActivateContextManager:
        def __init__(self, activate, deactivate):
            self.activate = activate
            self.deactivate = deactivate

        def __enter__(self):
            self.activate()
            return self

        def __exit__(self, type, value, traceback):
            self.deactivate()

        @classmethod
        def dummy(cls):
            def noop():
                pass

            return cls(noop, noop)

    class CompositeContextManager:
        def __init__(self, *managers):
            self.managers = managers

        def __enter__(self):
            for manager in self.managers:
                manager.__enter__()

        def __exit__(self, type, value, traceback):
            for manager in self.managers:
                manager.__exit__(type, value, traceback)

    def activate(self):
        if hasattr(self, '_mixin_activated'):
            if self._mixin_activated:
                raise ActivationError('This {} is already activated'.format(
                    type(self)))

        def do_activate():
            self._mixin_activated = True
            self._activate()

        def do_deactivate():
            self._deactivate()
            self._mixin_activated = False

        a = ActivateMixin.ActivateContextManager(do_activate, do_deactivate)
        subcomponents = self._subcomponents_to_activate()
        if subcomponents:
            return ActivateMixin.CompositeContextManager(
                a, ActivateMixin.compose(*subcomponents))
        else:
            return a

    @abstractmethod
    def _activate(self):
        pass

    @abstractmethod
    def _deactivate(self):
        pass

    def _subcomponents_to_activate(self):
        """Subclasses override this if they have subcomponents
        that may need to be activated when this class is activated
        """
        return []

    @staticmethod
    def with_activation(obj):
        """Method will give activate an object if it mixes in  the ActivateMixin and
        return the context manager, or else return a dummy context manager.
        """
        if obj is None or not isinstance(obj, ActivateMixin):
            return ActivateMixin.dummy()
        else:
            return obj.activate()

    @staticmethod
    def compose(*objs):
        managers = [
            obj.activate() for obj in objs
            if obj is not None and isinstance(obj, ActivateMixin)
        ]
        return ActivateMixin.CompositeContextManager(*managers)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/train.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rastervision.protos.tf_object_detection import optimizer_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_optimizer__pb2
from rastervision.protos.tf_object_detection import preprocessor_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_preprocessor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/train.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\n3rastervision/protos/tf_object_detection/train.proto\x12\'rastervision.protos.tf_object_detection\x1a\x37rastervision/protos/tf_object_detection/optimizer.proto\x1a:rastervision/protos/tf_object_detection/preprocessor.proto\"\x90\x08\n\x0bTrainConfig\x12\x16\n\nbatch_size\x18\x01 \x01(\r:\x02\x33\x32\x12]\n\x19\x64\x61ta_augmentation_options\x18\x02 \x03(\x0b\x32:.rastervision.protos.tf_object_detection.PreprocessingStep\x12\x1c\n\rsync_replicas\x18\x03 \x01(\x08:\x05\x66\x61lse\x12+\n\x1dkeep_checkpoint_every_n_hours\x18\x04 \x01(\r:\x04\x31\x30\x30\x30\x12\x45\n\toptimizer\x18\x05 \x01(\x0b\x32\x32.rastervision.protos.tf_object_detection.Optimizer\x12$\n\x19gradient_clipping_by_norm\x18\x06 \x01(\x02:\x01\x30\x12\x1e\n\x14\x66ine_tune_checkpoint\x18\x07 \x01(\t:\x00\x12#\n\x19\x66ine_tune_checkpoint_type\x18\x16 \x01(\t:\x00\x12,\n\x19\x66rom_detection_checkpoint\x18\x08 \x01(\x08:\x05\x66\x61lseB\x02\x18\x01\x12\x31\n\"load_all_detection_checkpoint_vars\x18\x13 \x01(\x08:\x05\x66\x61lse\x12\x14\n\tnum_steps\x18\t \x01(\r:\x01\x30\x12\x1f\n\x13startup_delay_steps\x18\n \x01(\x02:\x02\x31\x35\x12\x1f\n\x14\x62ias_grad_multiplier\x18\x0b \x01(\x02:\x01\x30\x12\"\n\x1aupdate_trainable_variables\x18\x19 \x03(\t\x12\x18\n\x10\x66reeze_variables\x18\x0c \x03(\t\x12 \n\x15replicas_to_aggregate\x18\r \x01(\x05:\x01\x31\x12!\n\x14\x62\x61tch_queue_capacity\x18\x0e \x01(\x05:\x03\x31\x35\x30\x12\"\n\x17num_batch_queue_threads\x18\x0f \x01(\x05:\x01\x38\x12\"\n\x17prefetch_queue_capacity\x18\x10 \x01(\x05:\x01\x35\x12)\n\x1amerge_multiple_label_boxes\x18\x11 \x01(\x08:\x05\x66\x61lse\x12$\n\x15use_multiclass_scores\x18\x18 \x01(\x08:\x05\x66\x61lse\x12%\n\x17\x61\x64\x64_regularization_loss\x18\x12 \x01(\x08:\x04true\x12$\n\x13max_number_of_boxes\x18\x14 \x01(\x05:\x03\x31\x30\x30\x42\x02\x18\x01\x12\'\n\x19unpad_groundtruth_tensors\x18\x15 \x01(\x08:\x04true\x12%\n\x16retain_original_images\x18\x17 \x01(\x08:\x05\x66\x61lse\x12\x1b\n\x0cuse_bfloat16\x18\x1a \x01(\x08:\x05\x66\x61lse')
  ,
  dependencies=[rastervision_dot_protos_dot_tf__object__detection_dot_optimizer__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_preprocessor__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_TRAINCONFIG = _descriptor.Descriptor(
  name='TrainConfig',
  full_name='rastervision.protos.tf_object_detection.TrainConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_size', full_name='rastervision.protos.tf_object_detection.TrainConfig.batch_size', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=32,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='data_augmentation_options', full_name='rastervision.protos.tf_object_detection.TrainConfig.data_augmentation_options', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sync_replicas', full_name='rastervision.protos.tf_object_detection.TrainConfig.sync_replicas', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='keep_checkpoint_every_n_hours', full_name='rastervision.protos.tf_object_detection.TrainConfig.keep_checkpoint_every_n_hours', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=1000,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='optimizer', full_name='rastervision.protos.tf_object_detection.TrainConfig.optimizer', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='gradient_clipping_by_norm', full_name='rastervision.protos.tf_object_detection.TrainConfig.gradient_clipping_by_norm', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fine_tune_checkpoint', full_name='rastervision.protos.tf_object_detection.TrainConfig.fine_tune_checkpoint', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fine_tune_checkpoint_type', full_name='rastervision.protos.tf_object_detection.TrainConfig.fine_tune_checkpoint_type', index=7,
      number=22, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='from_detection_checkpoint', full_name='rastervision.protos.tf_object_detection.TrainConfig.from_detection_checkpoint', index=8,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001'))),
    _descriptor.FieldDescriptor(
      name='load_all_detection_checkpoint_vars', full_name='rastervision.protos.tf_object_detection.TrainConfig.load_all_detection_checkpoint_vars', index=9,
      number=19, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_steps', full_name='rastervision.protos.tf_object_detection.TrainConfig.num_steps', index=10,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='startup_delay_steps', full_name='rastervision.protos.tf_object_detection.TrainConfig.startup_delay_steps', index=11,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(15),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bias_grad_multiplier', full_name='rastervision.protos.tf_object_detection.TrainConfig.bias_grad_multiplier', index=12,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='update_trainable_variables', full_name='rastervision.protos.tf_object_detection.TrainConfig.update_trainable_variables', index=13,
      number=25, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='freeze_variables', full_name='rastervision.protos.tf_object_detection.TrainConfig.freeze_variables', index=14,
      number=12, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='replicas_to_aggregate', full_name='rastervision.protos.tf_object_detection.TrainConfig.replicas_to_aggregate', index=15,
      number=13, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='batch_queue_capacity', full_name='rastervision.protos.tf_object_detection.TrainConfig.batch_queue_capacity', index=16,
      number=14, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=150,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_batch_queue_threads', full_name='rastervision.protos.tf_object_detection.TrainConfig.num_batch_queue_threads', index=17,
      number=15, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=8,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='prefetch_queue_capacity', full_name='rastervision.protos.tf_object_detection.TrainConfig.prefetch_queue_capacity', index=18,
      number=16, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=5,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='merge_multiple_label_boxes', full_name='rastervision.protos.tf_object_detection.TrainConfig.merge_multiple_label_boxes', index=19,
      number=17, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='use_multiclass_scores', full_name='rastervision.protos.tf_object_detection.TrainConfig.use_multiclass_scores', index=20,
      number=24, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='add_regularization_loss', full_name='rastervision.protos.tf_object_detection.TrainConfig.add_regularization_loss', index=21,
      number=18, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_number_of_boxes', full_name='rastervision.protos.tf_object_detection.TrainConfig.max_number_of_boxes', index=22,
      number=20, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=100,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=_descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001'))),
    _descriptor.FieldDescriptor(
      name='unpad_groundtruth_tensors', full_name='rastervision.protos.tf_object_detection.TrainConfig.unpad_groundtruth_tensors', index=23,
      number=21, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='retain_original_images', full_name='rastervision.protos.tf_object_detection.TrainConfig.retain_original_images', index=24,
      number=23, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='use_bfloat16', full_name='rastervision.protos.tf_object_detection.TrainConfig.use_bfloat16', index=25,
      number=26, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=214,
  serialized_end=1254,
)

_TRAINCONFIG.fields_by_name['data_augmentation_options'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_preprocessor__pb2._PREPROCESSINGSTEP
_TRAINCONFIG.fields_by_name['optimizer'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_optimizer__pb2._OPTIMIZER
DESCRIPTOR.message_types_by_name['TrainConfig'] = _TRAINCONFIG

TrainConfig = _reflection.GeneratedProtocolMessageType('TrainConfig', (_message.Message,), dict(
  DESCRIPTOR = _TRAINCONFIG,
  __module__ = 'rastervision.protos.tf_object_detection.train_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.TrainConfig)
  ))
_sym_db.RegisterMessage(TrainConfig)


_TRAINCONFIG.fields_by_name['from_detection_checkpoint'].has_options = True
_TRAINCONFIG.fields_by_name['from_detection_checkpoint']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001'))
_TRAINCONFIG.fields_by_name['max_number_of_boxes'].has_options = True
_TRAINCONFIG.fields_by_name['max_number_of_boxes']._options = _descriptor._ParseOptions(descriptor_pb2.FieldOptions(), _b('\030\001'))
# @@protoc_insertion_point(module_scope)



from abc import ABC, abstractmethod


class ChannelOrderError(Exception):
    def __init__(self, channel_order, num_channels):
        self.channel_order = channel_order
        self.num_channels = num_channels
        msg = 'The channel_order={} contains a channel index >= num_channels={}'
        super().__init__(msg.format(str(channel_order), num_channels))


class RasterSource(ABC):
    """A source of raster data.

    This should be subclassed when adding a new source of raster data such as
    a set of files, an API, a TMS URI schema, etc.
    """

    def __init__(self, channel_order, num_channels, raster_transformers=[]):
        """Construct a new RasterSource.

        Args:
            channel_order: list of channel indices to use when extracting chip from
                raw imagery.
            num_channels: Number of channels in the raw imagery before applying
                channel_order.
            raster_transformers: RasterTransformers used to transform chips
                whenever they are retrieved.
        """
        self.channel_order = channel_order
        self.num_channels = num_channels
        self.raster_transformers = raster_transformers

    def validate_channel_order(self, channel_order, num_channels):
        for c in channel_order:
            if c >= num_channels:
                raise ChannelOrderError(channel_order, num_channels)

    @abstractmethod
    def get_extent(self):
        """Return the extent of the RasterSource.

        Returns:
            Box in pixel coordinates with extent
        """
        pass

    @abstractmethod
    def get_dtype(self):
        """Return the numpy.dtype of this scene"""
        pass

    @abstractmethod
    def get_crs_transformer(self):
        """Return the associated CRSTransformer."""
        pass

    @abstractmethod
    def _get_chip(self, window):
        """Return the raw chip located in the window.

        Args:
            window: Box

        Returns:
            [height, width, channels] numpy array
        """
        pass

    def get_chip(self, window):
        """Return the transformed chip in the window.

        Get a raw chip, extract subset of channels using channel_order, and then apply
        transformations.

        Args:
            window: Box

        Returns:
            np.ndarray with shape [height, width, channels]
        """
        chip = self._get_chip(window)

        if self.channel_order:
            chip = chip[:, :, self.channel_order]

        for transformer in self.raster_transformers:
            chip = transformer.transform(chip, self.channel_order)

        return chip

    def get_raw_chip(self, window):
        """Return raw chip without using channel_order or applying transforms.

        Args:
            window: (Box) the window for which to get the chip

        Returns:
            np.ndarray with shape [height, width, channels]
        """
        return self._get_chip(window)

    def get_image_array(self):
        """Return entire transformed image array.

        Not safe to call on very large RasterSources.
        """
        return self.get_chip(self.get_extent())

    def get_raw_image_array(self):
        """Return entire raw image without using channel_order or applying transforms.

        Not safe to call on very large RasterSources.
        """
        return self.get_raw_chip(self.get_extent())



from torchvision import models


def get_model(model_arch, num_labels, pretrained=True):
    model = models.segmentation.segmentation._segm_resnet(
        'deeplabv3', 'resnet50', num_labels, False, pretrained_backbone=True)
    return model



from rastervision.data.label_source.utils import color_to_triple
import matplotlib
matplotlib.use('Agg')  # noqa


def plot_xy(ax, x, class_map, y=None):
    ax.axis('off')
    x = x.permute((1, 2, 0)).numpy()
    ax.imshow(x)

    if y is not None:
        colors = [class_map.get_by_id(i).color for i in range(len(class_map))]
        colors = [color_to_triple(c) for c in colors]
        colors = [tuple([_c / 255 for _c in c]) for c in colors]
        cmap = matplotlib.colors.ListedColormap(colors)

        y = y.numpy()
        ax.imshow(y, alpha=0.4, vmin=0, vmax=len(colors), cmap=cmap)



from rastervision.data.label import ObjectDetectionLabels
from rastervision.data.label_store import LabelStore
from rastervision.data.label_store.utils import boxes_to_geojson
from rastervision.data.vector_source import GeoJSONVectorSource
from rastervision.utils.files import json_to_file


class ObjectDetectionGeoJSONStore(LabelStore):
    def __init__(self, uri, crs_transformer, class_map):
        """Construct LabelStore backed by a GeoJSON file for object detection labels.

        Args:
            uri: uri of GeoJSON file containing labels
            crs_transformer: CRSTransformer to convert from map coords in label
                in GeoJSON file to pixel coords.
            class_map: ClassMap used to infer class_ids from class_name
                (or label) field
        """
        self.uri = uri
        self.crs_transformer = crs_transformer
        self.class_map = class_map

    def save(self, labels):
        """Save labels to URI."""
        boxes = labels.get_boxes()
        class_ids = labels.get_class_ids().tolist()
        scores = labels.get_scores().tolist()
        geojson = boxes_to_geojson(
            boxes,
            class_ids,
            self.crs_transformer,
            self.class_map,
            scores=scores)
        json_to_file(geojson, self.uri)

    def get_labels(self):
        vector_source = GeoJSONVectorSource(self.uri, self.crs_transformer)
        return ObjectDetectionLabels.from_geojson(vector_source.get_geojson())

    def empty_labels(self):
        return ObjectDetectionLabels.make_empty()



from os.path import join
import os
import shutil
import uuid
import logging

import numpy as np
from google.protobuf import json_format

from rastervision.backend import Backend
from rastervision.utils.files import (make_dir, get_local_path, upload_or_copy,
                                      download_if_needed, start_sync,
                                      sync_to_dir, sync_from_dir, list_paths)
from rastervision.utils.misc import save_img
from rastervision.data import ChipClassificationLabels

log = logging.getLogger(__name__)


def merge_class_dirs(scene_class_dirs, output_dir):
    seen_classes = set([])
    chip_ind = 0
    for scene_class_dir in scene_class_dirs:
        for class_name, src_class_dir in scene_class_dir.items():
            dst_class_dir = join(output_dir, class_name)
            if class_name not in seen_classes:
                make_dir(dst_class_dir)
                seen_classes.add(class_name)

            for src_class_file in [
                    join(src_class_dir, class_file)
                    for class_file in os.listdir(src_class_dir)
            ]:
                dst_class_file = join(dst_class_dir, '{}.png'.format(chip_ind))
                shutil.move(src_class_file, dst_class_file)
                chip_ind += 1


class FileGroup(object):
    def __init__(self, base_uri, tmp_dir):
        self.tmp_dir = tmp_dir
        self.base_uri = base_uri
        self.base_dir = self.get_local_path(base_uri)

        make_dir(self.base_dir)

    def get_local_path(self, uri):
        return get_local_path(uri, self.tmp_dir)

    def upload_or_copy(self, uri):
        upload_or_copy(self.get_local_path(uri), uri)

    def download_if_needed(self, uri):
        return download_if_needed(uri, self.tmp_dir)


class DatasetFiles(FileGroup):
    """Utilities for files produced when calling convert_training_data."""

    def __init__(self, base_uri, tmp_dir):
        FileGroup.__init__(self, base_uri, tmp_dir)

        self.partition_id = uuid.uuid4()

        self.training_zip_uri = join(
            base_uri, 'training-{}.zip'.format(self.partition_id))
        self.training_local_uri = join(self.base_dir,
                                       'training-{}'.format(self.partition_id))
        self.training_download_uri = self.get_local_path(
            join(self.base_uri, 'training'))
        make_dir(self.training_local_uri)

        self.validation_zip_uri = join(
            base_uri, 'validation-{}.zip'.format(self.partition_id))
        self.validation_local_uri = join(
            self.base_dir, 'validation-{}'.format(self.partition_id))
        self.validation_download_uri = self.get_local_path(
            join(self.base_uri, 'validation'))
        make_dir(self.validation_local_uri)

    def download(self):
        def _download(split, output_dir):
            scene_class_dirs = []
            for uri in list_paths(self.base_uri, 'zip'):
                base_name = os.path.basename(uri)
                if base_name.startswith(split):
                    data_zip_path = self.download_if_needed(uri)
                    data_dir = os.path.splitext(data_zip_path)[0]
                    shutil.unpack_archive(data_zip_path, data_dir)

                    # Append each of the directories containing this partitions'
                    # labeled images based on the class directory.
                    data_dir_subdirectories = next(os.walk(data_dir))[1]
                    scene_class_dirs.append(
                        dict([(class_name, os.path.join(data_dir, class_name))
                              for class_name in data_dir_subdirectories]))
            merge_class_dirs(scene_class_dirs, output_dir)

        _download('training', self.training_download_uri)
        _download('validation', self.validation_download_uri)

    def upload(self):
        def _upload(data_dir, zip_uri, split):
            if not any(os.scandir(data_dir)):
                log.warn(
                    'No data to write for split {} in partition {}...'.format(
                        split, self.partition_id))
            else:
                shutil.make_archive(data_dir, 'zip', data_dir)
                upload_or_copy(data_dir + '.zip', zip_uri)

        _upload(self.training_local_uri, self.training_zip_uri, 'training')
        _upload(self.validation_local_uri, self.validation_zip_uri,
                'validation')


class ModelFiles(FileGroup):
    """Utilities for files produced when calling train."""

    def __init__(self, base_uri, tmp_dir, replace_model=False):
        """Create these model files.

        Args:
            base_uri: Base URI of the model files
            replace_model: If the model file exists, remove.
                           Used for the training step, to retrain
                           existing models.

        Returns:
            A new ModelFile instance.
        """
        FileGroup.__init__(self, base_uri, tmp_dir)

        self.model_uri = join(self.base_uri, 'model')
        self.log_uri = join(self.base_uri, 'log.csv')

        if replace_model:
            if os.path.exists(self.model_uri):
                os.remove(self.model_uri)
            if os.path.exists(self.log_uri):
                os.remove(self.log_uri)

    def download_backend_config(self, pretrained_model_uri, kc_config,
                                dataset_files, class_map):
        from rastervision.protos.keras_classification.pipeline_pb2 \
            import PipelineConfig

        config = json_format.ParseDict(kc_config, PipelineConfig())

        # Update config using local paths.
        config.trainer.options.output_dir = self.get_local_path(self.base_uri)
        config.model.model_path = self.get_local_path(self.model_uri)
        config.model.nb_classes = len(class_map)

        config.trainer.options.training_data_dir = \
            dataset_files.training_download_uri
        config.trainer.options.validation_data_dir = \
            dataset_files.validation_download_uri

        del config.trainer.options.class_names[:]
        config.trainer.options.class_names.extend(class_map.get_class_names())

        # Save the pretrained weights locally
        pretrained_model_path = None
        if pretrained_model_uri:
            pretrained_model_path = self.download_if_needed(
                pretrained_model_uri)

        # Save an updated copy of the config file.
        config_path = os.path.join(self.tmp_dir, 'kc_config.json')
        config_str = json_format.MessageToJson(config)
        with open(config_path, 'w') as config_file:
            config_file.write(config_str)

        return (config_path, pretrained_model_path)


class KerasClassification(Backend):
    def __init__(self, backend_config, task_config):
        self.model = None
        self.config = backend_config
        self.class_map = task_config.class_map

    def process_scene_data(self, scene, data, tmp_dir):
        """Process each scene's training data

        Args:
            scene: Scene
            data: TrainingData

        Returns:
            dictionary of Scene's classes and corresponding local directory
                path
        """
        scratch_dir = join(tmp_dir, 'scratch-{}'.format(uuid.uuid4()))
        # Ensure directory is unique since scene id's could be shared between
        # training and test sets.
        scene_dir = join(scratch_dir, '{}-{}'.format(scene.id, uuid.uuid4()))
        class_dirs = {}

        for chip_idx, (chip, window, labels) in enumerate(data):
            class_id = labels.get_cell_class_id(window)
            # If a chip is not associated with a class, don't
            # use it in training data.
            if class_id is None:
                continue
            class_name = self.class_map.get_by_id(class_id).name
            class_dir = join(scene_dir, class_name)
            make_dir(class_dir)
            class_dirs[class_name] = class_dir
            chip_name = '{}.png'.format(chip_idx)
            chip_path = join(class_dir, chip_name)
            save_img(chip, chip_path)

        return class_dirs

    def process_sceneset_results(self, training_results, validation_results,
                                 tmp_dir):
        """After all scenes have been processed, collect all the images of
        each class across all scenes

        Args:
            training_results: list of dictionaries of training scenes'
                classes and corresponding local directory path
            validation_results: list of dictionaries of validation scenes'
                classes and corresponding local directory path
        """
        dataset_files = DatasetFiles(self.config.training_data_uri, tmp_dir)
        training_dir = dataset_files.training_local_uri
        validation_dir = dataset_files.validation_local_uri

        merge_class_dirs(training_results, training_dir)
        merge_class_dirs(validation_results, validation_dir)
        dataset_files.upload()

    def train(self, tmp_dir):
        from rastervision.backend.keras_classification.commands.train \
            import _train

        dataset_files = DatasetFiles(self.config.training_data_uri, tmp_dir)
        dataset_files.download()

        model_files = ModelFiles(
            self.config.training_output_uri,
            tmp_dir,
            replace_model=self.config.train_options.replace_model)
        model_paths = model_files.download_backend_config(
            self.config.pretrained_model_uri, self.config.kc_config,
            dataset_files, self.class_map)
        backend_config_path, pretrained_model_path = model_paths

        # Get output from potential previous run so we can resume training.
        if not self.config.train_options.replace_model:
            sync_from_dir(self.config.training_output_uri,
                          model_files.base_dir)

        sync = start_sync(
            model_files.base_dir,
            self.config.training_output_uri,
            sync_interval=self.config.train_options.sync_interval)
        with sync:
            do_monitoring = self.config.train_options.do_monitoring
            _train(backend_config_path, pretrained_model_path, do_monitoring)

        # Perform final sync
        sync_to_dir(
            model_files.base_dir, self.config.training_output_uri, delete=True)

    def load_model(self, tmp_dir):
        from rastervision.backend.keras_classification.builders \
            import model_builder

        if self.model is None:
            model_path = download_if_needed(self.config.model_uri, tmp_dir)
            self.model = model_builder.build_from_path(model_path)
            self.model._make_predict_function()

    def predict(self, chips, windows, tmp_dir):
        from rastervision.backend.keras_classification.utils \
            import predict

        # Ensure model is loaded
        self.load_model(tmp_dir)

        probs = predict(chips, self.model)

        labels = ChipClassificationLabels()

        for chip_probs, window in zip(probs, windows):
            # Add 1 to class_id since they start at 1.
            class_id = int(np.argmax(chip_probs) + 1)

            labels.set_cell(window, class_id, chip_probs)

        return labels



from os.path import join
from collections import defaultdict
import random
import glob

from PIL import Image
import torch
from torch.utils.data import Dataset, DataLoader
import torch.nn.functional as F
import torchvision

from rastervision.utils.files import (file_to_json)
from rastervision.backend.torch_utils.object_detection.boxlist import BoxList
from rastervision.backend.torch_utils.data import DataBunch


class ToTensor(object):
    def __init__(self):
        self.to_tensor = torchvision.transforms.ToTensor()

    def __call__(self, x, y):
        return (self.to_tensor(x), y)


class ScaleTransform(object):
    def __init__(self, height, width):
        self.height = height
        self.width = width

    def __call__(self, x, y):
        yscale = self.height / x.shape[1]
        xscale = self.width / x.shape[2]
        x = F.interpolate(
            x.unsqueeze(0), size=(self.height, self.width), mode='bilinear')[0]
        return (x, y.scale(yscale, xscale))


class RandomHorizontalFlip(object):
    def __init__(self, prob=0.5):
        self.prob = prob

    def __call__(self, x, y):
        if random.random() < self.prob:
            height, width = x.shape[-2:]
            x = x.flip(-1)

            boxes = y.boxes
            boxes[:, [1, 3]] = width - boxes[:, [3, 1]]
            y.boxes = boxes

        return (x, y)


class ComposeTransforms(object):
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, x, y):
        for t in self.transforms:
            x, y = t(x, y)
        return x, y


def collate_fn(data):
    x = [d[0].unsqueeze(0) for d in data]
    y = [d[1] for d in data]
    return (torch.cat(x), y)


class CocoDataset(Dataset):
    def __init__(self, img_dir, annotation_uris, transforms=None):
        self.img_dir = img_dir
        self.annotation_uris = annotation_uris
        self.transforms = transforms

        self.imgs = []
        self.img2id = {}
        self.id2img = {}
        self.id2boxes = defaultdict(lambda: [])
        self.id2labels = defaultdict(lambda: [])
        self.label2name = {}
        for annotation_uri in annotation_uris:
            ann_json = file_to_json(annotation_uri)
            for img in ann_json['images']:
                self.imgs.append(img['file_name'])
                self.img2id[img['file_name']] = img['id']
                self.id2img[img['id']] = img['file_name']
            for ann in ann_json['annotations']:
                img_id = ann['image_id']
                box = ann['bbox']
                label = ann['category_id']
                box = torch.tensor(
                    [[box[1], box[0], box[1] + box[3], box[0] + box[2]]])
                self.id2boxes[img_id].append(box)
                self.id2labels[img_id].append(label)
        self.id2boxes = dict([(id, torch.cat(boxes).float())
                              for id, boxes in self.id2boxes.items()])
        self.id2labels = dict([(id, torch.tensor(labels))
                               for id, labels in self.id2labels.items()])

    def __getitem__(self, ind):
        img_fn = self.imgs[ind]
        img_id = self.img2id[img_fn]
        img = Image.open(join(self.img_dir, img_fn))

        if img_id in self.id2boxes:
            boxes, labels = self.id2boxes[img_id], self.id2labels[img_id]
            boxlist = BoxList(boxes, labels=labels)
        else:
            boxlist = BoxList(
                torch.empty((0, 4)), labels=torch.empty((0, )).long())
        if self.transforms:
            return self.transforms(img, boxlist)
        return (img, boxlist)

    def __len__(self):
        return len(self.imgs)


def get_label_names(coco_path):
    categories = file_to_json(coco_path)['categories']
    label2name = dict([(cat['id'], cat['name']) for cat in categories])
    labels = ['background'
              ] + [label2name[i] for i in range(1,
                                                len(label2name) + 1)]
    return labels


def build_databunch(data_dir, img_sz, batch_sz):
    # TODO This is to avoid freezing in the middle of the first epoch. Would be nice
    # to fix this.
    num_workers = 0

    train_dir = join(data_dir, 'train')
    train_anns = glob.glob(join(train_dir, '*.json'))
    valid_dir = join(data_dir, 'valid')
    valid_anns = glob.glob(join(valid_dir, '*.json'))

    label_names = get_label_names(train_anns[0])
    aug_transforms = ComposeTransforms(
        [ToTensor(),
         ScaleTransform(img_sz, img_sz),
         RandomHorizontalFlip()])
    transforms = ComposeTransforms(
        [ToTensor(), ScaleTransform(img_sz, img_sz)])

    train_ds = CocoDataset(train_dir, train_anns, transforms=aug_transforms)
    valid_ds = CocoDataset(valid_dir, valid_anns, transforms=transforms)
    train_ds.label_names = label_names
    valid_ds.label_names = label_names

    train_dl = DataLoader(
        train_ds,
        shuffle=True,
        collate_fn=collate_fn,
        batch_size=batch_sz,
        num_workers=num_workers,
        drop_last=True,
        pin_memory=True)
    valid_dl = DataLoader(
        valid_ds,
        collate_fn=collate_fn,
        batch_size=batch_sz,
        num_workers=num_workers,
        pin_memory=True)
    return DataBunch(train_ds, train_dl, valid_ds, valid_dl, label_names)



from copy import deepcopy

import rastervision as rv
from rastervision.core.class_map import ClassMap
from rastervision.data.label_source import (LabelSourceConfig,
                                            LabelSourceConfigBuilder,
                                            SemanticSegmentationLabelSource)
from rastervision.protos.label_source_pb2 import LabelSourceConfig as LabelSourceConfigMsg
from rastervision.data.raster_source import RasterSourceConfig


class SemanticSegmentationLabelSourceConfig(LabelSourceConfig):
    def __init__(self, source, rgb_class_map=None):
        super().__init__(source_type=rv.SEMANTIC_SEGMENTATION_RASTER)
        self.source = source
        self.rgb_class_map = rgb_class_map

    def to_proto(self):
        msg = super().to_proto()

        rgb_class_items = None
        if self.rgb_class_map is not None:
            rgb_class_items = self.rgb_class_map.to_proto()
        opts = LabelSourceConfigMsg.SemanticSegmentationLabelSource(
            source=self.source.to_proto(), rgb_class_items=rgb_class_items)
        msg.semantic_segmentation_label_source.CopyFrom(opts)
        return msg

    def create_source(self, task_config, extent, crs_transformer, tmp_dir):
        raster_source = self.source.create_source(
            tmp_dir, crs_transformer, extent, task_config.class_map)
        return SemanticSegmentationLabelSource(raster_source,
                                               self.rgb_class_map)

    def update_for_command(self, command_type, experiment_config,
                           context=None):
        if context is None:
            context = []
        context = context + [self]
        self.source.update_for_command(command_type, experiment_config,
                                       context)

    def report_io(self, command_type, io_def):
        self.source.update_for_command(command_type, io_def)


class SemanticSegmentationLabelSourceConfigBuilder(LabelSourceConfigBuilder):
    def __init__(self, prev=None):
        config = {}
        if prev:
            config = {
                'source': prev.source,
                'rgb_class_map': prev.rgb_class_map
            }

        super().__init__(SemanticSegmentationLabelSourceConfig, config)

    def from_proto(self, msg):
        b = SemanticSegmentationLabelSourceConfigBuilder()

        label_source_msg = msg.semantic_segmentation_label_source
        # Add for backwards compatibility.
        if msg.HasField('semantic_segmentation_raster_source'):
            label_source_msg = msg.semantic_segmentation_raster_source

        raster_source_config = rv.RasterSourceConfig.from_proto(
            label_source_msg.source)

        b = self.with_raster_source(raster_source_config)
        rgb_class_items = msg.semantic_segmentation_raster_source.rgb_class_items

        b = b.with_raster_source(raster_source_config)
        rgb_class_items = label_source_msg.rgb_class_items

        if rgb_class_items:
            b = b.with_rgb_class_map(
                ClassMap.construct_from(list(rgb_class_items)))

        return b

    def with_raster_source(self, source, channel_order=None):
        """Set raster_source.

        Args:
            source: (RasterSourceConfig) A RasterSource assumed to have RGB values that
                are mapped to class_ids using the rgb_class_map.

        Returns:
            SemanticSegmentationLabelSourceConfigBuilder
        """
        b = deepcopy(self)
        if isinstance(source, RasterSourceConfig):
            b.config['source'] = source
        elif isinstance(source, str):
            provider = rv._registry.get_raster_source_default_provider(source)
            source = provider.construct(source, channel_order=channel_order)
            b.config['source'] = source
        else:
            raise rv.ConfigError(
                'source must be either string or RasterSourceConfig, '
                ' not {}'.format(str(type(source))))

        return b

    def with_rgb_class_map(self, rgb_class_map):
        """Set rgb_class_map.

        Args:
            rgb_class_map: (something accepted by ClassMap.construct_from) a class
                map with color values used to map RGB values to class ids

        Returns:
            SemanticSegmentationLabelSourceConfigBuilder
        """
        b = deepcopy(self)
        b.config['rgb_class_map'] = ClassMap.construct_from(rgb_class_map)
        return b

    def validate(self):
        source = self.config.get('source')

        if source is None:
            raise rv.ConfigError(
                'You must set the source for SemanticSegmentationLabelSourceConfig'
                ' Use "with_raster_source".')

        if not isinstance(source, (str, RasterSourceConfig)):
            raise rv.ConfigError(
                'raster_source must be of type str or RasterSourceConfig, got {}'.
                format(type(source)))



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/keras_classification/pipeline.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rastervision.protos.keras_classification import model_pb2 as rastervision_dot_protos_dot_keras__classification_dot_model__pb2
from rastervision.protos.keras_classification import trainer_pb2 as rastervision_dot_protos_dot_keras__classification_dot_trainer__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/keras_classification/pipeline.proto',
  package='keras_classification.protos',
  syntax='proto2',
  serialized_pb=_b('\n7rastervision/protos/keras_classification/pipeline.proto\x12\x1bkeras_classification.protos\x1a\x34rastervision/protos/keras_classification/model.proto\x1a\x36rastervision/protos/keras_classification/trainer.proto\"z\n\x0ePipelineConfig\x12\x31\n\x05model\x18\x01 \x02(\x0b\x32\".keras_classification.protos.Model\x12\x35\n\x07trainer\x18\x02 \x02(\x0b\x32$.keras_classification.protos.Trainer')
  ,
  dependencies=[rastervision_dot_protos_dot_keras__classification_dot_model__pb2.DESCRIPTOR,rastervision_dot_protos_dot_keras__classification_dot_trainer__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_PIPELINECONFIG = _descriptor.Descriptor(
  name='PipelineConfig',
  full_name='keras_classification.protos.PipelineConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='model', full_name='keras_classification.protos.PipelineConfig.model', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='trainer', full_name='keras_classification.protos.PipelineConfig.trainer', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=198,
  serialized_end=320,
)

_PIPELINECONFIG.fields_by_name['model'].message_type = rastervision_dot_protos_dot_keras__classification_dot_model__pb2._MODEL
_PIPELINECONFIG.fields_by_name['trainer'].message_type = rastervision_dot_protos_dot_keras__classification_dot_trainer__pb2._TRAINER
DESCRIPTOR.message_types_by_name['PipelineConfig'] = _PIPELINECONFIG

PipelineConfig = _reflection.GeneratedProtocolMessageType('PipelineConfig', (_message.Message,), dict(
  DESCRIPTOR = _PIPELINECONFIG,
  __module__ = 'rastervision.protos.keras_classification.pipeline_pb2'
  # @@protoc_insertion_point(class_scope:keras_classification.protos.PipelineConfig)
  ))
_sym_db.RegisterMessage(PipelineConfig)


# @@protoc_insertion_point(module_scope)



from abc import (ABC, abstractmethod)
import os
import inspect

from rastervision.utils.files import download_or_copy


class ConfigError(Exception):
    pass


class Config(ABC):
    @abstractmethod
    def to_builder(self):
        """Return a builder based on this config.
        """
        pass  # pragma: no cover

    @abstractmethod
    def to_proto(self):
        """Returns the protobuf configuration for this config.
        """
        pass  # pragma: no cover

    def update_for_command(self,
                           command_type,
                           experiment_config,
                           context=None,
                           io_def=None):
        """Updates this configuration for the given command

        Note: While configuration is immutable for client facing operations,
        this is an internal operation and mutates the configuration.

        Args:
            command_type: The command type that is currently being
                preprocessed. experiment_config: The experiment configuration
                that this configuration is a part of.
            context: Optional list of parent configurations, to allow for child
                configurations contained in collections to understand their
                context in the experiment configuration.

        Returns:
          Nothing. Call should mutate the configuration object itself.
        """
        pass  # pragma: no cover

    @abstractmethod
    def report_io(self, command_type, io_def):
        """Updates the given CommandIODefinition.

        So that it includes the inputs, outputs, and missing files for this
        configuration at this command.

        Args:
            command_type: The command type that is currently being preprocessed.
            io_def: The CommandIODefinition that this call should modify.

        Returns: Nothing. This call should make the appropriate calls to the
            given io_def to mutate its state.
        """
        pass

    @staticmethod
    @abstractmethod
    def builder():
        """Returns a new builder that takes this configuration
           as its starting point.
        """
        pass  # pragma: no cover

    @staticmethod
    @abstractmethod
    def from_proto(msg):
        """Creates a Config from the specificed protobuf message
        TODO: Allow loading from file uri or dict
        """
        pass  # pragma: no cover


class ConfigBuilder(ABC):
    def __init__(self, config_class, config=None):
        """Construct a builder.

           Args:
             config_class: The Config class that this builder builds.
             config: A dictionary of **kwargs that will eventually be passed
                     into the __init__ method of config_class to build the configuration.
                     This config is modified with the fluent builder methods.
        """
        if config is None:  # pragma: no cover
            config = {}

        self.config_class = config_class
        self.config = config

    def build(self):
        """Returns the configuration that is built by this builder.
        """
        self.validate()
        arguments = set(inspect.getfullargspec(self.config_class).args)
        keys = set(self.config.keys())
        config = {k: self.config[k] for k in (arguments & keys)}
        return self.config_class(**config)

    def validate(self):
        """Validate this config, if there is validation on the builder that
           is not captured by the required arguments of the config.
        """
        pass  # pragma: no cover

    @abstractmethod
    def from_proto(self, msg):
        """Return a builder that takes the configuration from the proto message
           as its starting point.
        """
        pass  # pragma: no cover


class BundledConfigMixin(ABC):
    """Mixin for configurations that participate in the bundling of a
    prediction package"""

    @abstractmethod
    def save_bundle_files(self, bundle_dir):
        """Place files into a bundle directory for bundling into
        a prediction package.

        Returns: A tuple of (config, uris) of the modified configuration
                 with the basenames of URIs in place of the original URIs,
                 and a list of URIs that are to be bundled.
        """
        pass  # pragma: no cover

    def bundle_file(self, uri, bundle_dir):
        local_path = download_or_copy(uri, bundle_dir)
        base_name = os.path.basename(local_path)
        return (local_path, base_name)

    @abstractmethod
    def load_bundle_files(self, bundle_dir):
        """Load files from a prediction package bundle directory."""
        pass  # pragma: no cover



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: rastervision/protos/tf_object_detection/ssd.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from rastervision.protos.tf_object_detection import anchor_generator_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_anchor__generator__pb2
from rastervision.protos.tf_object_detection import box_coder_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_box__coder__pb2
from rastervision.protos.tf_object_detection import box_predictor_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_box__predictor__pb2
from rastervision.protos.tf_object_detection import hyperparams_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_hyperparams__pb2
from rastervision.protos.tf_object_detection import image_resizer_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_image__resizer__pb2
from rastervision.protos.tf_object_detection import matcher_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_matcher__pb2
from rastervision.protos.tf_object_detection import losses_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_losses__pb2
from rastervision.protos.tf_object_detection import post_processing_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_post__processing__pb2
from rastervision.protos.tf_object_detection import region_similarity_calculator_pb2 as rastervision_dot_protos_dot_tf__object__detection_dot_region__similarity__calculator__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='rastervision/protos/tf_object_detection/ssd.proto',
  package='rastervision.protos.tf_object_detection',
  syntax='proto2',
  serialized_pb=_b('\n1rastervision/protos/tf_object_detection/ssd.proto\x12\'rastervision.protos.tf_object_detection\x1a>rastervision/protos/tf_object_detection/anchor_generator.proto\x1a\x37rastervision/protos/tf_object_detection/box_coder.proto\x1a;rastervision/protos/tf_object_detection/box_predictor.proto\x1a\x39rastervision/protos/tf_object_detection/hyperparams.proto\x1a;rastervision/protos/tf_object_detection/image_resizer.proto\x1a\x35rastervision/protos/tf_object_detection/matcher.proto\x1a\x34rastervision/protos/tf_object_detection/losses.proto\x1a=rastervision/protos/tf_object_detection/post_processing.proto\x1aJrastervision/protos/tf_object_detection/region_similarity_calculator.proto\"\x94\t\n\x03Ssd\x12\x13\n\x0bnum_classes\x18\x01 \x01(\x05\x12L\n\rimage_resizer\x18\x02 \x01(\x0b\x32\x35.rastervision.protos.tf_object_detection.ImageResizer\x12W\n\x11\x66\x65\x61ture_extractor\x18\x03 \x01(\x0b\x32<.rastervision.protos.tf_object_detection.SsdFeatureExtractor\x12\x44\n\tbox_coder\x18\x04 \x01(\x0b\x32\x31.rastervision.protos.tf_object_detection.BoxCoder\x12\x41\n\x07matcher\x18\x05 \x01(\x0b\x32\x30.rastervision.protos.tf_object_detection.Matcher\x12\x62\n\x15similarity_calculator\x18\x06 \x01(\x0b\x32\x43.rastervision.protos.tf_object_detection.RegionSimilarityCalculator\x12)\n\x1a\x65ncode_background_as_zeros\x18\x0c \x01(\x08:\x05\x66\x61lse\x12 \n\x15negative_class_weight\x18\r \x01(\x02:\x01\x31\x12L\n\rbox_predictor\x18\x07 \x01(\x0b\x32\x35.rastervision.protos.tf_object_detection.BoxPredictor\x12R\n\x10\x61nchor_generator\x18\x08 \x01(\x0b\x32\x38.rastervision.protos.tf_object_detection.AnchorGenerator\x12P\n\x0fpost_processing\x18\t \x01(\x0b\x32\x37.rastervision.protos.tf_object_detection.PostProcessing\x12+\n\x1dnormalize_loss_by_num_matches\x18\n \x01(\x08:\x04true\x12-\n\x1enormalize_loc_loss_by_codesize\x18\x0e \x01(\x08:\x05\x66\x61lse\x12;\n\x04loss\x18\x0b \x01(\x0b\x32-.rastervision.protos.tf_object_detection.Loss\x12\x1f\n\x10\x66reeze_batchnorm\x18\x10 \x01(\x08:\x05\x66\x61lse\x12\'\n\x18inplace_batchnorm_update\x18\x0f \x01(\x08:\x05\x66\x61lse\x12.\n\x1fweight_regression_loss_by_score\x18\x11 \x01(\x08:\x05\x66\x61lse\x12>\n/use_expected_classification_loss_under_sampling\x18\x12 \x01(\x08:\x05\x66\x61lse\x12$\n\x19minimum_negative_sampling\x18\x13 \x01(\x02:\x01\x30\x12*\n\x1f\x64\x65sired_negative_sampling_ratio\x18\x14 \x01(\x02:\x01\x33\"\x96\x03\n\x13SsdFeatureExtractor\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x1b\n\x10\x64\x65pth_multiplier\x18\x02 \x01(\x02:\x01\x31\x12\x15\n\tmin_depth\x18\x03 \x01(\x05:\x02\x31\x36\x12N\n\x10\x63onv_hyperparams\x18\x04 \x01(\x0b\x32\x34.rastervision.protos.tf_object_detection.Hyperparams\x12:\n+override_base_feature_extractor_hyperparams\x18\t \x01(\x08:\x05\x66\x61lse\x12\x1a\n\x0fpad_to_multiple\x18\x05 \x01(\x05:\x01\x31\x12#\n\x14use_explicit_padding\x18\x07 \x01(\x08:\x05\x66\x61lse\x12\x1c\n\ruse_depthwise\x18\x08 \x01(\x08:\x05\x66\x61lse\x12L\n\x03\x66pn\x18\n \x01(\x0b\x32?.rastervision.protos.tf_object_detection.FeaturePyramidNetworksJ\x04\x08\x06\x10\x07\"i\n\x16\x46\x65\x61turePyramidNetworks\x12\x14\n\tmin_level\x18\x01 \x01(\x05:\x01\x33\x12\x14\n\tmax_level\x18\x02 \x01(\x05:\x01\x37\x12#\n\x16\x61\x64\x64itional_layer_depth\x18\x03 \x01(\x05:\x03\x32\x35\x36')
  ,
  dependencies=[rastervision_dot_protos_dot_tf__object__detection_dot_anchor__generator__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_box__coder__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_box__predictor__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_hyperparams__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_image__resizer__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_matcher__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_losses__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_post__processing__pb2.DESCRIPTOR,rastervision_dot_protos_dot_tf__object__detection_dot_region__similarity__calculator__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_SSD = _descriptor.Descriptor(
  name='Ssd',
  full_name='rastervision.protos.tf_object_detection.Ssd',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_classes', full_name='rastervision.protos.tf_object_detection.Ssd.num_classes', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='image_resizer', full_name='rastervision.protos.tf_object_detection.Ssd.image_resizer', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='feature_extractor', full_name='rastervision.protos.tf_object_detection.Ssd.feature_extractor', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='box_coder', full_name='rastervision.protos.tf_object_detection.Ssd.box_coder', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='matcher', full_name='rastervision.protos.tf_object_detection.Ssd.matcher', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='similarity_calculator', full_name='rastervision.protos.tf_object_detection.Ssd.similarity_calculator', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='encode_background_as_zeros', full_name='rastervision.protos.tf_object_detection.Ssd.encode_background_as_zeros', index=6,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='negative_class_weight', full_name='rastervision.protos.tf_object_detection.Ssd.negative_class_weight', index=7,
      number=13, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='box_predictor', full_name='rastervision.protos.tf_object_detection.Ssd.box_predictor', index=8,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='anchor_generator', full_name='rastervision.protos.tf_object_detection.Ssd.anchor_generator', index=9,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='post_processing', full_name='rastervision.protos.tf_object_detection.Ssd.post_processing', index=10,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='normalize_loss_by_num_matches', full_name='rastervision.protos.tf_object_detection.Ssd.normalize_loss_by_num_matches', index=11,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='normalize_loc_loss_by_codesize', full_name='rastervision.protos.tf_object_detection.Ssd.normalize_loc_loss_by_codesize', index=12,
      number=14, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='loss', full_name='rastervision.protos.tf_object_detection.Ssd.loss', index=13,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='freeze_batchnorm', full_name='rastervision.protos.tf_object_detection.Ssd.freeze_batchnorm', index=14,
      number=16, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='inplace_batchnorm_update', full_name='rastervision.protos.tf_object_detection.Ssd.inplace_batchnorm_update', index=15,
      number=15, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='weight_regression_loss_by_score', full_name='rastervision.protos.tf_object_detection.Ssd.weight_regression_loss_by_score', index=16,
      number=17, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='use_expected_classification_loss_under_sampling', full_name='rastervision.protos.tf_object_detection.Ssd.use_expected_classification_loss_under_sampling', index=17,
      number=18, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='minimum_negative_sampling', full_name='rastervision.protos.tf_object_detection.Ssd.minimum_negative_sampling', index=18,
      number=19, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='desired_negative_sampling_ratio', full_name='rastervision.protos.tf_object_detection.Ssd.desired_negative_sampling_ratio', index=19,
      number=20, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(3),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=645,
  serialized_end=1817,
)


_SSDFEATUREEXTRACTOR = _descriptor.Descriptor(
  name='SsdFeatureExtractor',
  full_name='rastervision.protos.tf_object_detection.SsdFeatureExtractor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='rastervision.protos.tf_object_detection.SsdFeatureExtractor.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='depth_multiplier', full_name='rastervision.protos.tf_object_detection.SsdFeatureExtractor.depth_multiplier', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=True, default_value=float(1),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='min_depth', full_name='rastervision.protos.tf_object_detection.SsdFeatureExtractor.min_depth', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=16,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conv_hyperparams', full_name='rastervision.protos.tf_object_detection.SsdFeatureExtractor.conv_hyperparams', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='override_base_feature_extractor_hyperparams', full_name='rastervision.protos.tf_object_detection.SsdFeatureExtractor.override_base_feature_extractor_hyperparams', index=4,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pad_to_multiple', full_name='rastervision.protos.tf_object_detection.SsdFeatureExtractor.pad_to_multiple', index=5,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=1,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='use_explicit_padding', full_name='rastervision.protos.tf_object_detection.SsdFeatureExtractor.use_explicit_padding', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='use_depthwise', full_name='rastervision.protos.tf_object_detection.SsdFeatureExtractor.use_depthwise', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='fpn', full_name='rastervision.protos.tf_object_detection.SsdFeatureExtractor.fpn', index=8,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1820,
  serialized_end=2226,
)


_FEATUREPYRAMIDNETWORKS = _descriptor.Descriptor(
  name='FeaturePyramidNetworks',
  full_name='rastervision.protos.tf_object_detection.FeaturePyramidNetworks',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='min_level', full_name='rastervision.protos.tf_object_detection.FeaturePyramidNetworks.min_level', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=3,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='max_level', full_name='rastervision.protos.tf_object_detection.FeaturePyramidNetworks.max_level', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=7,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='additional_layer_depth', full_name='rastervision.protos.tf_object_detection.FeaturePyramidNetworks.additional_layer_depth', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=256,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2228,
  serialized_end=2333,
)

_SSD.fields_by_name['image_resizer'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_image__resizer__pb2._IMAGERESIZER
_SSD.fields_by_name['feature_extractor'].message_type = _SSDFEATUREEXTRACTOR
_SSD.fields_by_name['box_coder'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_box__coder__pb2._BOXCODER
_SSD.fields_by_name['matcher'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_matcher__pb2._MATCHER
_SSD.fields_by_name['similarity_calculator'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_region__similarity__calculator__pb2._REGIONSIMILARITYCALCULATOR
_SSD.fields_by_name['box_predictor'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_box__predictor__pb2._BOXPREDICTOR
_SSD.fields_by_name['anchor_generator'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_anchor__generator__pb2._ANCHORGENERATOR
_SSD.fields_by_name['post_processing'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_post__processing__pb2._POSTPROCESSING
_SSD.fields_by_name['loss'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_losses__pb2._LOSS
_SSDFEATUREEXTRACTOR.fields_by_name['conv_hyperparams'].message_type = rastervision_dot_protos_dot_tf__object__detection_dot_hyperparams__pb2._HYPERPARAMS
_SSDFEATUREEXTRACTOR.fields_by_name['fpn'].message_type = _FEATUREPYRAMIDNETWORKS
DESCRIPTOR.message_types_by_name['Ssd'] = _SSD
DESCRIPTOR.message_types_by_name['SsdFeatureExtractor'] = _SSDFEATUREEXTRACTOR
DESCRIPTOR.message_types_by_name['FeaturePyramidNetworks'] = _FEATUREPYRAMIDNETWORKS

Ssd = _reflection.GeneratedProtocolMessageType('Ssd', (_message.Message,), dict(
  DESCRIPTOR = _SSD,
  __module__ = 'rastervision.protos.tf_object_detection.ssd_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.Ssd)
  ))
_sym_db.RegisterMessage(Ssd)

SsdFeatureExtractor = _reflection.GeneratedProtocolMessageType('SsdFeatureExtractor', (_message.Message,), dict(
  DESCRIPTOR = _SSDFEATUREEXTRACTOR,
  __module__ = 'rastervision.protos.tf_object_detection.ssd_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.SsdFeatureExtractor)
  ))
_sym_db.RegisterMessage(SsdFeatureExtractor)

FeaturePyramidNetworks = _reflection.GeneratedProtocolMessageType('FeaturePyramidNetworks', (_message.Message,), dict(
  DESCRIPTOR = _FEATUREPYRAMIDNETWORKS,
  __module__ = 'rastervision.protos.tf_object_detection.ssd_pb2'
  # @@protoc_insertion_point(class_scope:rastervision.protos.tf_object_detection.FeaturePyramidNetworks)
  ))
_sym_db.RegisterMessage(FeaturePyramidNetworks)


# @@protoc_insertion_point(module_scope)



import rastervision as rv
from rastervision.data.label import ObjectDetectionLabels
from rastervision.data.label_source import LabelSource


class ObjectDetectionLabelSource(LabelSource):
    def __init__(self, vector_source, crs_transformer, class_map, extent):
        """Constructor.

        Args:
            vector_source: (VectorSource or str)
            crs_transformer: CRSTransformer to convert from map coords in label
                in GeoJSON file to pixel coords.
            class_map: ClassMap used to infer class_ids from class_name
                (or label) field
            extent: Box used to filter the labels by extent
        """
        if isinstance(vector_source, str):
            provider = rv._registry.get_vector_source_default_provider(
                vector_source)
            vector_source = provider.construct(vector_source) \
                .create_source(
                    crs_transformer=crs_transformer, extent=extent, class_map=class_map)

        self.labels = ObjectDetectionLabels.from_geojson(
            vector_source.get_geojson(), extent=extent)

    def get_labels(self, window=None):
        if window is None:
            return self.labels

        return ObjectDetectionLabels.get_overlapping(self.labels, window)



from copy import deepcopy
from typing import (List, Dict, Tuple, Union)

import rastervision as rv
from rastervision.task import ObjectDetection
from rastervision.core.class_map import (ClassMap, ClassItem)
from rastervision.task import (TaskConfig, TaskConfigBuilder)
from rastervision.protos.task_pb2 import TaskConfig as TaskConfigMsg
from rastervision.protos.class_item_pb2 import ClassItem as ClassItemMsg


class ObjectDetectionConfig(TaskConfig):
    class ChipOptions:
        def __init__(self,
                     neg_ratio=1,
                     ioa_thresh=0.8,
                     window_method='chip',
                     label_buffer=0.0):
            self.neg_ratio = neg_ratio
            self.ioa_thresh = ioa_thresh
            self.window_method = window_method
            self.label_buffer = label_buffer

    class PredictOptions:
        def __init__(self, merge_thresh=0.5, score_thresh=0.5):
            self.merge_thresh = merge_thresh
            self.score_thresh = score_thresh

    def __init__(self,
                 class_map,
                 predict_batch_size=10,
                 predict_package_uri=None,
                 debug=True,
                 predict_debug_uri=None,
                 chip_size=300,
                 chip_options=ChipOptions(),
                 predict_options=PredictOptions()):
        super().__init__(rv.OBJECT_DETECTION, predict_batch_size,
                         predict_package_uri, debug, predict_debug_uri)
        self.class_map = class_map
        self.chip_size = chip_size
        self.chip_options = chip_options
        self.predict_options = predict_options

    def create_task(self, backend):
        return ObjectDetection(self, backend)

    def to_proto(self):
        msg = super().to_proto()
        chip_options = TaskConfigMsg.ObjectDetectionConfig.ChipOptions(
            neg_ratio=self.chip_options.neg_ratio,
            ioa_thresh=self.chip_options.ioa_thresh,
            window_method=self.chip_options.window_method,
            label_buffer=self.chip_options.label_buffer)

        predict_options = TaskConfigMsg.ObjectDetectionConfig.PredictOptions(
            merge_thresh=self.predict_options.merge_thresh,
            score_thresh=self.predict_options.score_thresh)

        conf = TaskConfigMsg.ObjectDetectionConfig(
            chip_size=self.chip_size,
            class_items=self.class_map.to_proto(),
            chip_options=chip_options,
            predict_options=predict_options)
        msg.MergeFrom(
            TaskConfigMsg(
                object_detection_config=conf,
                predict_package_uri=self.predict_package_uri))

        return msg

    def save_bundle_files(self, bundle_dir):
        return (self, [])

    def load_bundle_files(self, bundle_dir):
        return self


class ObjectDetectionConfigBuilder(TaskConfigBuilder):
    def __init__(self, prev=None):
        config = {}
        if prev:
            config = {
                'predict_batch_size': prev.predict_batch_size,
                'predict_package_uri': prev.predict_package_uri,
                'debug': prev.debug,
                'predict_debug_uri': prev.predict_debug_uri,
                'class_map': prev.class_map,
                'chip_size': prev.chip_size,
                'chip_options': prev.chip_options,
                'predict_options': prev.predict_options
            }
        super().__init__(ObjectDetectionConfig, config)

    def validate(self):
        if 'class_map' not in self.config:
            raise rv.ConfigError('Class map required for this task. '
                                 'Use "with_classes"')
        if not isinstance(self.config['class_map'], ClassMap):
            raise rv.ConfigError(
                'Class map set with "with_classes" must be of type ClassMap, got {}'.
                format(type(self.config['class_map'])))

    def from_proto(self, msg):
        b = super().from_proto(msg)
        conf = msg.object_detection_config

        return b.with_classes(list(conf.class_items)) \
                .with_predict_package_uri(msg.predict_package_uri) \
                .with_chip_size(conf.chip_size) \
                .with_chip_options(neg_ratio=conf.chip_options.neg_ratio,
                                   ioa_thresh=conf.chip_options.ioa_thresh,
                                   window_method=conf.chip_options.window_method,
                                   label_buffer=conf.chip_options.label_buffer) \
                .with_predict_options(merge_thresh=conf.predict_options.merge_thresh,
                                      score_thresh=conf.predict_options.score_thresh)

    def with_classes(
            self, classes: Union[ClassMap, List[str], List[ClassItemMsg], List[
                ClassItem], Dict[str, int], Dict[str, Tuple[int, str]]]):
        """Set the classes for this task.

            Args:
                classes: Either a list of class names, a dict which
                         maps class names to class ids, or a dict
                         which maps class names to a tuple of (class_id, color),
                         where color is a PIL color string.
        """
        b = deepcopy(self)
        b.config['class_map'] = ClassMap.construct_from(classes)
        return b

    def with_chip_size(self, chip_size):
        """Set the chip_size for this task.

        Note that some model implementations have a minimum size of input they
        can handle. A value of > 200 is usually safe.

        Args:
            chip_size: (int) chip size in units of pixels
        """
        b = deepcopy(self)
        b.config['chip_size'] = chip_size
        return b

    def with_chip_options(self,
                          neg_ratio=1,
                          ioa_thresh=0.8,
                          window_method='chip',
                          label_buffer=0.0):
        """Sets object detection configurations for the Chip command

        Args:
            neg_ratio: The ratio of negative chips (those containing no bounding
                boxes) to positive chips. This can be useful if the statistics
                of the background is different in positive chips. For example,
                in car detection, the positive chips will always contain roads,
                but no examples of rooftops since cars tend to not be near
                rooftops. This option is not used when window_method is
                `sliding`.

            ioa_thresh: When a box is partially outside of a training chip, it
                is not clear if (a clipped version) of the box should be
                included in the chip. If the IOA (intersection over area) of the
                box with the chip is greater than ioa_thresh, it is included in
                the chip.

            window_method: Different models in the Object Detection API have
                different inputs. Some models allow variable size inputs so
                several methods of building training data are required

                Valid values are:
                - chip (default)
                - label
                   - each label's bounding box is the positive window
                - image
                   - each image is the positive window
                - sliding
                   - each image is from a sliding window with 50% overlap

            label_buffer:
                If method is "label", the positive window can be buffered.
                If value is >= 0. and < 1., the value is treated as a percentage
                If value is >= 1., the value is treated in number of pixels
        """
        b = deepcopy(self)
        b.config['chip_options'] = ObjectDetectionConfig.ChipOptions(
            neg_ratio=neg_ratio,
            ioa_thresh=ioa_thresh,
            window_method=window_method,
            label_buffer=label_buffer)
        return b

    def with_predict_options(self, merge_thresh=0.5, score_thresh=0.5):
        """Prediction options for this task.

        Args:
           merge_thresh: If predicted boxes have an IOA (intersection over area)
                         greater than merge_thresh, then they are merged into a
                         single box during postprocessing. This is needed since
                         the sliding window approach results in some false
                         duplicates.

           score_thresh: Predicted boxes are only output if their score is
                         above score_thresh.
        """
        b = deepcopy(self)
        b.config['predict_options'] = ObjectDetectionConfig.PredictOptions(
            merge_thresh=merge_thresh, score_thresh=score_thresh)
        return b



from rastervision.data import ActivateMixin


class Scene(ActivateMixin):
    """The raster data and labels associated with an area of interest."""

    def __init__(self,
                 id,
                 raster_source,
                 ground_truth_label_source=None,
                 prediction_label_store=None,
                 aoi_polygons=None):
        """Construct a new Scene.

        Args:
            id: ID for this scene
            raster_source: RasterSource for this scene
            ground_truth_label_store: optional LabelSource
            prediction_label_store: optional LabelStore
            aoi: Optional list of AOI polygons
        """
        self.id = id
        self.raster_source = raster_source
        self.ground_truth_label_source = ground_truth_label_source
        self.prediction_label_store = prediction_label_store
        if aoi_polygons is None:
            self.aoi_polygons = []
        else:
            self.aoi_polygons = aoi_polygons

    def _subcomponents_to_activate(self):
        return [
            self.raster_source, self.ground_truth_label_source,
            self.prediction_label_store
        ]

    def _activate(self):
        pass

    def _deactivate(self):
        pass



# Copied from https://github.com/developmentseed/label-maker/blob/master/label_maker/filter.py
# flake8: noqa

# pylint: disable=eval-used,too-many-return-statements
"""Create a feature filtering function from a Mapbox GL Filter."""

# Python port of https://github.com/mapbox/mapbox-gl-js/blob/c9900db279db776f493ce8b6749966cedc2d6b8a/src/style-spec/feature_filter/index.js


def create_filter(filt):
    """Create a feature filtering function from a Mapbox GL Filter.

    Given a filter expressed as nested lists, return a new function
    that evaluates whether a given feature (with a .properties or .tags property)
    passes its test. More information:
    - https://www.mapbox.com/mapbox-gl-js/style-spec/#other-filter
    - https://github.com/mapbox/mapbox-gl-js/tree/master/src/style-spec/feature_filter

    Parameters
    ------------
    filt: list
        Mapbox GL filter

    Returns
    --------
    func: function
        A function which evaluates whether a GeoJSON feature meets the input filter criteria
    """

    def func(f):
        """evaluates whether a given feature passes its filter"""
        p = f.get('properties', {}) if f else {}  # pylint: disable=unused-variable
        return eval(_compile(filt))

    return func


def _compile(filt):
    """Return a string represented the compiled filter function"""
    if not filt:
        return 'True'
    op = filt[0]
    if len(filt) == 1:
        return 'False' if op == 'any' else 'True'
    if op in ['==', '!=', '<', '>', '<=', '>=']:
        return _compile_comparison_op(filt[1], filt[2], op)
    elif op == 'any':
        return _compile_logical_op(filt[1:], ' or ')
    elif op == 'all':
        return _compile_logical_op(filt[1:], ' and ')
    elif op == 'none':
        return _compile_negation(_compile_logical_op(filt[1:], ' or '))
    elif op == 'in':
        return _compile_in_op(filt[1], filt[2:])
    elif op == '!in':
        return _compile_negation(_compile_in_op(filt[1], filt[2:]))
    elif op == 'has':
        return _compile_has_op(filt[1])
    elif op == '!has':
        return _compile_negation(_compile_has_op(filt[1]))
    return 'True'


def _compile_property_reference(prop):
    """Find the correct reference on the input feature"""
    if prop == '$type':
        return 'f.get("geometry").get("type")'
    elif prop == '$id':
        return 'f.get("id")'
    return 'p.get("{}")'.format(prop)


def _compile_comparison_op(prop, value, op):
    """Combine two values with a comparison operator"""
    left = _compile_property_reference(prop)
    right = _stringify(value)
    return left + op + right


def _compile_logical_op(expressions, op):
    """Join multiple logical expressions"""
    return op.join(map(_compile, expressions))


def _compile_in_op(prop, values):
    """Test if a property is within a list of values"""
    return '{} in {}'.format(_compile_property_reference(prop), values)


def _compile_has_op(prop):
    """Test if a property exists on a feature"""
    return '"id" in f' if prop == '$id' else '{} in p'.format(_stringify(prop))


def _compile_negation(expression):
    """Negate the input expression"""
    return 'not ({})'.format(expression)


def _stringify(s):
    """Convert input to string, wrap with quotes if already a string"""
    return '"{}"'.format(s) if isinstance(s, str) else str(s)



import numpy as np

from rastervision.data.raster_transformer.raster_transformer \
    import RasterTransformer


class StatsTransformer(RasterTransformer):
    """Transforms non-uint8 to uint8 values using raster_stats.
    """

    def __init__(self, raster_stats=None):
        """Construct a new StatsTransformer.

        Args:
            raster_stats: (RasterStats) used to transform chip to have
                desired statistics
        """
        self.raster_stats = raster_stats

    def transform(self, chip, channel_order=None):
        """Transform a chip.

        Transforms non-uint8 to uint8 values using raster_stats.

        Args:
            chip: ndarray of shape [height, width, channels] This is assumed to already
                have the channel_order applied to it if channel_order is set. In other
                words, channels should be equal to len(channel_order).
            channel_order: list of indices of channels that were extracted from the
                raw imagery.

        Returns:
            [height, width, channels] uint8 numpy array

        """
        if chip.dtype != np.uint8:
            if self.raster_stats:
                if channel_order is None:
                    channel_order = np.arange(chip.shape[2])

                # Subtract mean and divide by std to get zscores.
                means = np.array(self.raster_stats.means)
                means = means[np.newaxis, np.newaxis, channel_order].astype(
                    np.float)
                stds = np.array(self.raster_stats.stds)
                stds = stds[np.newaxis, np.newaxis, channel_order].astype(
                    np.float)

                # Don't transform NODATA zero values.
                nodata = chip == 0

                chip = chip - means
                chip = chip / stds

                # Make zscores that fall between -3 and 3 span 0 to 255.
                chip += 3
                chip /= 6

                chip = np.clip(chip, 0, 1)
                chip *= 255
                chip = chip.astype(np.uint8)

                chip[nodata] = 0
            else:
                raise ValueError('raster_stats not defined.')

        return chip
